"""
Tests for species and gene additions in v3.23.0.

Covers:
- Notamacropus rufogriseus (wallaby) with Maru prefix
- Kryptolebias marmoratus killifish class I genes (A1-H1) and Krma prefix
- Additional Ctenomys tuco-tuco species
- Batch rodent species (Dipodomys, Octodon, Heterocephalus, etc.)
- Batch fish species (Merluccius, Liparis, Oncorhynchus nerka, etc.)
- New bird/reptile/amphibian species
"""

import pytest

from mhcgnomes import Gene, Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# All new species with their prefixes and optional alt-prefixes
# ---------------------------------------------------------------------------

NEW_SPECIES = [
    # Wallaby
    ("NotaRufo", "Notamacropus rufogriseus", ["Maru"]),
    # Ctenomys
    ("CtenHaig", "Ctenomys haigi", []),
    ("CtenArge", "Ctenomys argentinus", []),
    ("CtenFrat", "Ctenomys frater", []),
    ("CtenOpim", "Ctenomys opimus", []),
    # Rodents
    ("DipoOrdi", "Dipodomys ordii", []),
    ("OctoDegu", "Octodon degus", []),
    ("HeteGlab", "Heterocephalus glaber", []),
    ("NannGali", "Nannospalax galili", []),
    ("PeroMani", "Peromyscus maniculatus", []),
    ("PeroLeuc", "Peromyscus leucopus", []),
    ("JacuJacu", "Jaculus jaculus", []),
    ("SigmHisp", "Sigmodon hispidus", []),
    ("CricBara", "Cricetulus barabensis", []),
    ("CynoGunn", "Cynomys gunnisoni", []),
    ("DeloSubl", "Delomys sublineatus", []),
    ("RattFusc", "Rattus fuscipes", []),
    # Mus species
    ("MusMinu", "Mus minutoides", []),
    ("MusPaha", "Mus pahari", []),
    ("MusSpic", "Mus spicilegus", []),
    ("MusShor", "Mus shortridgei", []),
    ("MusGrat", "Mus gratus", []),
    ("MusSetu", "Mus setulosus", []),
    ("MusTerr", "Mus terricolor", []),
    # Fish
    ("MerlPoll", "Merluccius polli", ["Mepo"]),
    ("LipaTana", "Liparis tanakae", []),
    ("TripTibe", "Triplophysa tibetana", []),
    ("AuloHans", "Aulonocara hansbaenschi", []),
    ("OdonPota", "Odontobutis potamophilus", []),
    ("HippErec", "Hippocampus erectus", []),
    ("OncoNerk", "Oncorhynchus nerka", []),
    ("PolySpat", "Polyodon spathula", []),
    # Bird / Reptile / Amphibian
    ("HaliAlbi", "Haliaeetus albicilla", []),
    ("TiliRugo", "Tiliqua rugosa", []),
    ("ConoSubc", "Conolophus subcristatus", []),
    ("EspaPros", "Espadarana prosoblepon", []),
]


@pytest.mark.parametrize(
    "prefix, latin, alt_prefixes",
    NEW_SPECIES,
    ids=[s[0] for s in NEW_SPECIES],
)
def test_new_species_parseable(prefix, latin, alt_prefixes):
    species = Species.get(prefix)
    assert species is not None, f"Species not found for prefix {prefix}"
    assert species.latin_name == latin


@pytest.mark.parametrize(
    "prefix, latin, alt_prefixes",
    NEW_SPECIES,
    ids=[s[0] for s in NEW_SPECIES],
)
def test_new_species_latin_name_lookup(prefix, latin, alt_prefixes):
    species = Species.get(latin)
    assert species is not None, f"Species not found for latin name {latin}"


@pytest.mark.parametrize(
    "prefix, latin, alt_prefixes",
    [s for s in NEW_SPECIES if s[2]],
    ids=[s[0] for s in NEW_SPECIES if s[2]],
)
def test_new_species_alt_prefix(prefix, latin, alt_prefixes):
    for alt in alt_prefixes:
        species = Species.get(alt)
        assert species is not None, f"Alt prefix {alt} not found for {latin}"
        assert species.latin_name == latin


# ---------------------------------------------------------------------------
# Wallaby: Notamacropus rufogriseus
# ---------------------------------------------------------------------------


def test_maru_DAB():
    eq_(parse("Maru-DAB"), Gene.get("Maru", "DAB"))


def test_maru_UA():
    eq_(parse("Maru-UA"), Gene.get("Maru", "UA"))


def test_maru_DBB():
    eq_(parse("Maru-DBB"), Gene.get("Maru", "DBB"))


# ---------------------------------------------------------------------------
# Killifish: Kryptolebias marmoratus class I genes
# ---------------------------------------------------------------------------


def test_krma_A1():
    eq_(parse("Krma-A1"), Gene.get("Krma", "A1"))


def test_krma_B1():
    eq_(parse("Krma-B1"), Gene.get("Krma", "B1"))


def test_krma_C1():
    eq_(parse("Krma-C1"), Gene.get("Krma", "C1"))


def test_krypmarm_H1():
    eq_(parse("KrypMarm-H1"), Gene.get("KrypMarm", "H1"))


def test_krma_inherits_fish_UA():
    """Killifish should still inherit UA from Actinopterygii sp."""
    gene = Gene.get("Krma", "UA")
    assert gene is not None


# ---------------------------------------------------------------------------
# Ctenomys: new species
# ---------------------------------------------------------------------------


def test_ctenomys_haigi_DRB():
    eq_(parse("CtenHaig-DRB"), Gene.get("CtenHaig", "DRB"))


def test_ctenomys_opimus_DQB():
    eq_(parse("CtenOpim-DQB"), Gene.get("CtenOpim", "DQB"))


# ---------------------------------------------------------------------------
# Rodents: gene inheritance
# ---------------------------------------------------------------------------


def test_dipodomys_inherits_DRB1():
    eq_(parse("DipoOrdi-DRB1"), Gene.get("DipoOrdi", "DRB1"))


def test_heterocephalus_inherits_DQB1():
    eq_(parse("HeteGlab-DQB1"), Gene.get("HeteGlab", "DQB1"))


def test_peromyscus_inherits_DRB1():
    eq_(parse("PeroMani-DRB1"), Gene.get("PeroMani", "DRB1"))


def test_mus_minutoides_inherits_K():
    """Mus species should inherit H2 genes from Mus sp."""
    eq_(parse("MusMinu-K"), Gene.get("MusMinu", "K"))


def test_mus_pahari_inherits_D():
    eq_(parse("MusPaha-D"), Gene.get("MusPaha", "D"))


def test_mus_spicilegus_inherits_EB():
    eq_(parse("MusSpic-EB"), Gene.get("MusSpic", "EB"))


def test_rattus_fuscipes_inherits_A():
    eq_(parse("RattFusc-A"), Gene.get("RattFusc", "A"))


# ---------------------------------------------------------------------------
# Fish: gene inheritance
# ---------------------------------------------------------------------------


def test_merluccius_polli_mepo_prefix():
    eq_(parse("Mepo-UA"), Gene.get("Mepo", "UA"))


def test_oncorhynchus_nerka_inherits_salmonid_UAA():
    eq_(parse("OncoNerk-UAA"), Gene.get("OncoNerk", "UAA"))


def test_polyodon_inherits_DAB():
    eq_(parse("PolySpat-DAB"), Gene.get("PolySpat", "DAB"))


# ---------------------------------------------------------------------------
# Bird / Reptile / Amphibian: gene inheritance
# ---------------------------------------------------------------------------


def test_haliaeetus_inherits_DAB1():
    eq_(parse("HaliAlbi-DAB1"), Gene.get("HaliAlbi", "DAB1"))


def test_tiliqua_inherits_UAA():
    eq_(parse("TiliRugo-UAA"), Gene.get("TiliRugo", "UAA"))


def test_conolophus_inherits_UA():
    eq_(parse("ConoSubc-UA"), Gene.get("ConoSubc", "UA"))


def test_espadarana_inherits_UA():
    eq_(parse("EspaPros-UA"), Gene.get("EspaPros", "UA"))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Optional, Union

from .allele import Allele
from .class2_locus import Class2Locus
from .common import cache
from .errors import ParseError
from .gene import Gene
from .gene_class_info import GeneClassInfo
from .mhc_class import MhcClass
from .pair import Pair
from .parser import (
    COLLAPSE_SINGLETON_HAPLOTYPES,
    COLLAPSE_SINGLETON_SEROTYPES,
    DEFAULT_SPECIES_PREFIX,
    GENE_SEPS,
    INFER_CLASS2_PAIRING,
    USE_ALLELE_ALIASES,
    Parser,
)
from .result import Result
from .species import Species, infer_species_from_context_only_prefix


@cache
def cached_parser(
    use_allele_aliases: bool = USE_ALLELE_ALIASES,
    gene_seps=GENE_SEPS,
    collapse_singleton_haplotypes: bool = COLLAPSE_SINGLETON_HAPLOTYPES,
    collapse_singleton_serotypes: bool = COLLAPSE_SINGLETON_SEROTYPES,
    verbose: bool = False,
) -> Parser:
    """
    Get or create a cached Parser instance.

    Construct a Parser instance if this combination of arguments hasn't
    been used before, otherwise retrieve an existing parser from cache.

    Parameters
    ----------
    use_allele_aliases : bool
        Whether to use allele alias mappings for parsing.
    gene_seps : tuple of str
        Separator characters between gene name and allele fields.
    collapse_singleton_haplotypes : bool
        If a Haplotype contains only a single allele, return the allele
        instead of a haplotype.
    collapse_singleton_serotypes : bool
        If a Serotype contains only a single allele, return the allele
        instead of a serotype.
    verbose : bool
        Print intermediate parsing steps for debugging.

    Returns
    -------
    Parser
        A Parser instance configured with the given options.

    Examples
    --------
    >>> parser = cached_parser()
    >>> result = parser.parse("HLA-A*02:01")
    >>> result.gene_name
    'A'
    """
    return Parser(
        use_allele_aliases=use_allele_aliases,
        gene_seps=gene_seps,
        collapse_singleton_haplotypes=collapse_singleton_haplotypes,
        collapse_singleton_serotypes=collapse_singleton_serotypes,
        verbose=verbose,
    )


def _reparse_result_for_species(
    parser: Parser,
    result: Result,
    expected_species: Species,
    infer_class2_pairing: bool,
    required_result_types,
    preferred_result_types,
    only_class1: bool,
    only_class2: bool,
    max_allele_fields,
):
    if type(result) is Species or not hasattr(result, "species"):
        return None

    speciesless_string = result.to_string(include_species=False)
    reparsed = parser.parse(
        f"{expected_species.prefix}-{speciesless_string}",
        default_species=expected_species,
        infer_class2_pairing=infer_class2_pairing,
        raise_on_error=False,
        required_result_types=required_result_types,
        preferred_result_types=preferred_result_types,
        only_class1=only_class1,
        only_class2=only_class2,
        max_allele_fields=max_allele_fields,
    )
    if reparsed is None or type(reparsed) is not type(result):
        return None
    if not hasattr(reparsed, "species") or reparsed.species != expected_species:
        return None
    return reparsed.copy(raw_string=result.raw_string)


def _reparse_with_context_only_prefix(
    parser: Parser,
    raw_string: str,
    expected_species: Species,
    infer_class2_pairing: bool,
    required_result_types,
    preferred_result_types,
    only_class1: bool,
    only_class2: bool,
    max_allele_fields,
):
    inferred = infer_species_from_context_only_prefix(raw_string)
    if inferred is None:
        return None
    candidate_species, matched_prefix = inferred
    if expected_species not in candidate_species:
        return None
    rewritten = f"{expected_species.prefix}{raw_string[len(matched_prefix) :]}"
    reparsed = parser.parse(
        rewritten,
        default_species=expected_species,
        infer_class2_pairing=infer_class2_pairing,
        raise_on_error=False,
        required_result_types=required_result_types,
        preferred_result_types=preferred_result_types,
        only_class1=only_class1,
        only_class2=only_class2,
        max_allele_fields=max_allele_fields,
    )
    if reparsed is None:
        return None
    return reparsed.copy(raw_string=raw_string)


def _format_species_options(species_objects, max_items=4):
    names = [sp.name for sp in species_objects]
    shown = names[:max_items]
    if len(names) > max_items:
        shown.append(f"+{len(names) - max_items} more")
    return ", ".join(shown)


def _format_canonical_prefix_options(species_objects, max_items=4):
    shown = [f"{sp.prefix} ({sp.name})" for sp in species_objects[:max_items]]
    if len(species_objects) > max_items:
        shown.append(f"+{len(species_objects) - max_items} more")
    return ", ".join(shown)


def _context_only_prefix_error_message(raw_string: str):
    inferred = infer_species_from_context_only_prefix(raw_string)
    if inferred is None:
        return None

    candidate_species, matched_prefix = inferred
    candidate_species = list(candidate_species)
    canonical_options = _format_canonical_prefix_options(candidate_species)
    global_owner = Species.get(matched_prefix)

    if global_owner is not None:
        return (
            f"Prefix '{matched_prefix}' is also used in source data for "
            f"{_format_species_options(candidate_species)}. "
            f"Runtime keeps '{matched_prefix}' for '{global_owner.name}'. "
            f"Use species='<latin name>' or a collision-free canonical prefix such as "
            f"{canonical_options}."
        )

    return (
        f"Prefix '{matched_prefix}' is source-attested for multiple species "
        f"({_format_species_options(candidate_species)}) and is not globally unique. "
        f"Use species='<latin name>' or a collision-free canonical prefix such as "
        f"{canonical_options}."
    )


_NON_MHC_REGION_GENE_NAMES = {
    "ABCB2": "TAP1",
    "ABCB3": "TAP2",
    "B2M": "B2M",
    "CIITA": "CIITA",
    "HAM1": "TAP1",
    "HAM2": "TAP2",
    "HM13": "HM13",
    "PRR3": "PRR3",
    "PSF1": "TAP1",
    "PSF2": "TAP2",
    "RING11": "TAP2",
    "RING4": "TAP1",
    "TAP-L": "TAP-L",
    "TAPL": "TAP-L",
    "TAP1": "TAP1",
    "TAP2": "TAP2",
    "TAPBP": "TAPBP",
}

_LENIENT_GENE_SUFFIX_RULES = {
    "A-U": ("I", "alpha"),
    "BLB": ("IIa", "beta"),
    "DAA": ("IIa", "alpha"),
    "DAB": ("IIa", "beta"),
    "DMA": ("IIb", "alpha"),
    "DMB": ("IIb", "beta"),
    "DNB": ("IIa", "beta"),
    "DPA1": ("IIa", "alpha"),
    "DQB1": ("IIa", "beta"),
    "DRA": ("IIa", "alpha"),
    "DR-1": ("IIa", "alpha"),
    "DXB": ("IIa", "beta"),
    "E-S": ("I", "alpha"),
    "F10": ("I", "alpha"),
    "M": ("IIa", "alpha"),
    "M1": ("IIa", "beta"),
    "Q10": ("I", "alpha"),
    "Q9": ("I", "alpha"),
    "U": ("I", "alpha"),
    "UA": ("I", "alpha"),
    "UAA": ("I", "alpha"),
    "UB": ("I", "alpha"),
}

_LENIENT_GENE_SUFFIX_PATTERNS = (
    (re.compile(r"^DAB\d+$", re.IGNORECASE), ("IIa", "beta")),
    (re.compile(r"^DRB\d+(?:-\d+)?$", re.IGNORECASE), ("IIa", "beta")),
)


def _chain_for_species_gene(species: Species, gene_name: str, mhc_class: Optional[str]):
    if mhc_class is None or mhc_class == "other":
        return None
    if mhc_class in {"I", "Ia", "Ib", "Ic", "Id"}:
        return "alpha"
    return species.class2_gene_name_to_chain_type.get(gene_name)


def _build_gene_class_info_from_parsed(parsed: Result, raw_string: str):
    if parsed is None:
        return None

    if isinstance(parsed, Allele):
        gene_name = parsed.gene.name
        mhc_class = parsed.mhc_class
        return GeneClassInfo(
            species=parsed.species,
            gene_name=gene_name,
            mhc_class=mhc_class,
            chain=_chain_for_species_gene(parsed.species, gene_name, mhc_class),
            non_mhc=mhc_class == "other",
            source="parsed_allele",
            raw_string=raw_string,
        )

    if isinstance(parsed, Gene):
        gene_name = parsed.name
        mhc_class = parsed.mhc_class
        return GeneClassInfo(
            species=parsed.species,
            gene_name=gene_name,
            mhc_class=mhc_class,
            chain=_chain_for_species_gene(parsed.species, gene_name, mhc_class),
            non_mhc=mhc_class == "other",
            source="parsed_gene",
            raw_string=raw_string,
        )

    if isinstance(parsed, Class2Locus):
        return GeneClassInfo(
            species=parsed.species,
            gene_name=parsed.name,
            mhc_class=parsed.mhc_class,
            chain=None,
            non_mhc=False,
            source="parsed_locus",
            raw_string=raw_string,
        )

    if isinstance(parsed, Pair):
        return GeneClassInfo(
            species=parsed.species,
            gene_name=f"{parsed.alpha.gene_name}/{parsed.beta.gene_name}",
            mhc_class=parsed.mhc_class,
            chain=None,
            non_mhc=False,
            source="parsed_pair",
            raw_string=raw_string,
        )

    if isinstance(parsed, MhcClass):
        return GeneClassInfo(
            species=parsed.species,
            gene_name=None,
            mhc_class=parsed.mhc_class,
            chain=parsed.chain,
            non_mhc=parsed.mhc_class == "other",
            source="parsed_class",
            raw_string=raw_string,
        )

    return None


def _normalize_inference_gene_token(gene_token: str):
    return gene_token.strip().strip(",;").upper()


def _strip_gene_token(candidate_gene_string: str):
    candidate_gene_string = candidate_gene_string.strip()
    if not candidate_gene_string:
        return None
    if any(ch.isspace() for ch in candidate_gene_string):
        return None
    candidate_gene_string = candidate_gene_string.split("/", 1)[0]
    candidate_gene_string = candidate_gene_string.split("*", 1)[0]
    candidate_gene_string = candidate_gene_string.split("#", 1)[0]
    candidate_gene_string = candidate_gene_string.strip(" -_.:;")
    if not candidate_gene_string:
        return None
    return candidate_gene_string


def _extract_species_and_gene_string(
    raw_string: str,
    parser: Parser,
    expected_species: Optional[Species],
):
    explicit_species, remaining = parser.parse_species_from_prefix(raw_string)
    if explicit_species is not None:
        if expected_species is not None and explicit_species != expected_species:
            return (
                None,
                None,
                (
                    f"Input prefix selects species '{explicit_species.name}', "
                    f"not requested species '{expected_species.name}'."
                ),
            )
        return explicit_species if expected_species is None else expected_species, remaining, None

    inferred = infer_species_from_context_only_prefix(raw_string)
    if inferred is not None:
        candidate_species, matched_prefix = inferred
        if expected_species is None:
            return None, None, _context_only_prefix_error_message(raw_string)
        if expected_species not in candidate_species:
            return (
                None,
                None,
                (
                    f"Prefix '{matched_prefix}' is not source-attested for "
                    f"'{expected_species.name}'."
                ),
            )
        remaining = raw_string[len(matched_prefix) :]
        return expected_species, remaining, None

    if expected_species is not None:
        return expected_species, raw_string, None

    return None, raw_string, None


def _infer_gene_rule(gene_token: str):
    normalized = _normalize_inference_gene_token(gene_token)
    if normalized in _LENIENT_GENE_SUFFIX_RULES:
        mhc_class, chain = _LENIENT_GENE_SUFFIX_RULES[normalized]
        return normalized, mhc_class, chain
    for regex, (mhc_class, chain) in _LENIENT_GENE_SUFFIX_PATTERNS:
        if regex.fullmatch(normalized):
            return normalized, mhc_class, chain
    return None


def _build_gene_class_info_from_family(species: Species, gene_token: str, raw_string: str):
    family_name = species.find_matching_gene_family_name(gene_token)
    if family_name is None:
        return None

    member_names = species.get_gene_family_members(family_name)
    member_classes = {species.get_mhc_class_of_gene(member_name) for member_name in member_names}
    if len(member_classes) != 1:
        return None
    mhc_class = member_classes.pop()

    member_chains = {
        _chain_for_species_gene(species, member_name, mhc_class) for member_name in member_names
    }
    chain = member_chains.pop() if len(member_chains) == 1 else None
    return GeneClassInfo(
        species=species,
        gene_name=family_name,
        mhc_class=mhc_class,
        chain=chain,
        non_mhc=mhc_class == "other",
        source="ontology_family",
        raw_string=raw_string,
    )


def _build_gene_class_info_from_token(species: Species, gene_token: str, raw_string: str):
    canonical_gene_name = species.find_matching_gene_name(gene_token)
    if canonical_gene_name is not None:
        mhc_class = species.get_mhc_class_of_gene(canonical_gene_name)
        return GeneClassInfo(
            species=species,
            gene_name=canonical_gene_name,
            mhc_class=mhc_class,
            chain=_chain_for_species_gene(species, canonical_gene_name, mhc_class),
            non_mhc=mhc_class == "other",
            source="canonical_gene",
            raw_string=raw_string,
        )

    locus = Class2Locus.get(species, gene_token)
    if locus is not None:
        return GeneClassInfo(
            species=species,
            gene_name=locus.name,
            mhc_class=locus.mhc_class,
            chain=None,
            non_mhc=False,
            source="canonical_locus",
            raw_string=raw_string,
        )

    family_info = _build_gene_class_info_from_family(species, gene_token, raw_string)
    if family_info is not None:
        return family_info

    normalized = _normalize_inference_gene_token(gene_token)
    if normalized in _NON_MHC_REGION_GENE_NAMES:
        return GeneClassInfo(
            species=species,
            gene_name=_NON_MHC_REGION_GENE_NAMES[normalized],
            mhc_class="other",
            chain=None,
            non_mhc=True,
            source="non_mhc_name",
            raw_string=raw_string,
        )

    inferred = _infer_gene_rule(gene_token)
    if inferred is not None:
        normalized_name, mhc_class, chain = inferred
        return GeneClassInfo(
            species=species,
            gene_name=normalized_name,
            mhc_class=mhc_class,
            chain=chain,
            non_mhc=False,
            source="heuristic_suffix",
            raw_string=raw_string,
        )

    return None


def parse_gene_class(
    raw_string: str,
    default_species=DEFAULT_SPECIES_PREFIX,
    species: Union[str, None] = None,
    raise_on_error: bool = True,
) -> Optional[GeneClassInfo]:
    """
    Leniently infer the MHC class/chain for gene-like inputs.

    This function first attempts the normal strict parser. If that fails, it
    falls back to species-aware gene-token inference for source-backed suffixes
    such as ``F10`` or ``DR-1`` and for known non-MHC-region helper genes like
    ``CIITA`` or ``TAP1``. Exact ontology-backed family names such as ``TAP``
    can be classified without inventing a specific member gene.
    """
    expected_species = Species.get(species) if species is not None else None
    if species is not None and expected_species is None:
        if raise_on_error:
            raise ParseError(f"Unknown species '{species}'")
        return None

    parsed = parse(
        raw_string,
        default_species=default_species,
        species=species,
        raise_on_error=False,
    )
    info = _build_gene_class_info_from_parsed(parsed, raw_string)
    if info is not None:
        return info

    parser = cached_parser()
    inferred_species, candidate_gene_string, error_message = _extract_species_and_gene_string(
        raw_string=raw_string,
        parser=parser,
        expected_species=expected_species,
    )
    if error_message is not None:
        if raise_on_error:
            raise ParseError(error_message)
        return None

    # Heuristic inference should only run when the caller supplied a strict
    # species or when the raw string carried an explicit species prefix.
    if inferred_species is None:
        if raise_on_error:
            raise ParseError(f"Could not infer MHC class from '{raw_string}'")
        return None

    candidate_gene_string = parser.strip_extra_chars(candidate_gene_string)
    gene_token = _strip_gene_token(candidate_gene_string)
    if gene_token is None:
        if raise_on_error:
            raise ParseError(f"Could not infer MHC class from '{raw_string}'")
        return None

    info = _build_gene_class_info_from_token(inferred_species, gene_token, raw_string)
    if info is not None:
        return info

    if raise_on_error:
        raise ParseError(f"Could not infer MHC class from '{raw_string}'")
    return None


def infer_mhc_class(
    raw_string: str,
    default_species=DEFAULT_SPECIES_PREFIX,
    species: Union[str, None] = None,
) -> Optional[str]:
    """
    Convenience wrapper around :func:`parse_gene_class` returning only the
    inferred MHC class string.
    """
    result = parse_gene_class(
        raw_string,
        default_species=default_species,
        species=species,
        raise_on_error=False,
    )
    if result is None:
        return None
    return result.mhc_class


def parse(
    raw_string: str,
    default_species=DEFAULT_SPECIES_PREFIX,
    species: Union[str, None] = None,
    use_allele_aliases: bool = USE_ALLELE_ALIASES,
    infer_class2_pairing: bool = INFER_CLASS2_PAIRING,
    collapse_singleton_haplotypes: bool = COLLAPSE_SINGLETON_HAPLOTYPES,
    collapse_singleton_serotypes: bool = COLLAPSE_SINGLETON_SEROTYPES,
    max_allele_fields=None,
    required_result_types=None,
    preferred_result_types=None,
    only_class1: bool = False,
    only_class2: bool = False,
    verbose: bool = False,
    raise_on_error: bool = True,
    require_explicit_species: bool = False,
) -> Optional[Result]:
    """
    Parse MHC alleles into a structured representation.

    Parameters
    ----------
    raw_string : str
       String corresponding to allele, locus, or other MHC-related name

    default_species : str
       By default, parse alleles like "A*02:01" as human but it's possible
       to change this to some other species.

    species : str or None
       Strict species constraint. If provided, the final parse result must have
       this species exactly. Unlike default_species, this does not allow the
       parser to fall back to a different species when the input is explicit
       or ambiguous. If a non-Species result is first parsed at a generic
       ancestor taxon, it may be reparsed for the requested descendant species
       when that conversion is valid.

    use_allele_aliases : bool

    infer_class2_pairing : bool
       If given only the alpha or beta chain of a Class II allele,
       try to infer the most likely pairing from population frequencies.

    collapse_singleton_haplotypes : bool
        If a Haplotype contains only a single allele or Class II allele pair,
        then return the allele instead of a haplotype.

    collapse_singleton_serotypes : bool
        If a Serotype contains only a single allele or Class II allele pair,
        then return the allele instead of a serotype.

    max_allele_fields : int
        If not None, restrict number of allele fields to given value.

    required_result_types : list of type
        Strict filter. Only return results of the given classes.

    preferred_result_types : list of type
        Prefer returning one of these classes when available.
        If none match, return the best non-preferred parse.

    only_class1 : bool
        Only return MHC Class I results

    only_class2 : bool
        Only return MHC Class II results

    verbose : bool
        Print intermediate parsing steps

    raise_on_error : bool
        Raise an exception if string can't be parsed. If False, return None
        instead.

    require_explicit_species : bool
        Only accept a result whose species the input actually named. When the
        species was inferred from a gene name, or came from default_species,
        treat the string as unparsed. Useful for validating curated data,
        where a confident but inferred species is worse than no answer.
        See Result.species_source and Result.species_from_input.

    Returns
    -------
    Result
        A parsed MHC object, which may be an Allele, AlleleWithoutGene, Gene,
        Species, Haplotype, Serotype, Supertype, Pair, Class2Locus, or MhcClass
        depending on the input string. Returns None if parsing fails and
        raise_on_error is False.

    Raises
    ------
    ParseError
        If the string cannot be parsed and raise_on_error is True.

    Examples
    --------
    >>> from mhcgnomes import parse
    >>> allele = parse("HLA-A*02:01")
    >>> allele.gene_name
    'A'
    >>> allele.allele_fields
    ('02', '01')
    >>> allele.to_string()
    'HLA-A*02:01'
    """
    if preferred_result_types is None:
        preferred_result_types = []
    if required_result_types is None:
        required_result_types = []
    parser = cached_parser(
        use_allele_aliases=use_allele_aliases,
        collapse_singleton_haplotypes=collapse_singleton_haplotypes,
        collapse_singleton_serotypes=collapse_singleton_serotypes,
        verbose=verbose,
    )

    # species= is the strict form: the final result must have this species.
    # default_species= is the lenient form: used only when the string
    # doesn't specify a species.
    # They are mutually exclusive.
    if species is not None and default_species != DEFAULT_SPECIES_PREFIX:
        raise ValueError("Cannot specify both 'species' and 'default_species'")

    expected_species = Species.get(species) if species is not None else None
    effective_default = species if species is not None else default_species

    parser_error = None
    try:
        result = parser.parse(
            raw_string,
            default_species=effective_default,
            infer_class2_pairing=infer_class2_pairing,
            raise_on_error=raise_on_error,
            required_result_types=required_result_types,
            preferred_result_types=preferred_result_types,
            only_class1=only_class1,
            only_class2=only_class2,
            max_allele_fields=max_allele_fields,
            require_explicit_species=require_explicit_species,
        )
    except ParseError as error:
        parser_error = error
        result = None

    if species is not None and expected_species is not None:
        rescued = _reparse_with_context_only_prefix(
            parser=parser,
            raw_string=raw_string,
            expected_species=expected_species,
            infer_class2_pairing=infer_class2_pairing,
            required_result_types=required_result_types,
            preferred_result_types=preferred_result_types,
            only_class1=only_class1,
            only_class2=only_class2,
            max_allele_fields=max_allele_fields,
        )
        if rescued is not None:
            result = rescued
            parser_error = None

    if parser_error is not None:
        maybe_context_message = (
            _context_only_prefix_error_message(raw_string) if species is None else None
        )
        if maybe_context_message is not None:
            raise ParseError(maybe_context_message) from parser_error
        raise parser_error

    # Strict species check: if species= was provided, the final result must
    # have that exact species.
    #
    # Rules:
    # - Species results: exact match only.
    # - Non-Species results with .species: allow ancestor→descendant conversion
    #   only when reparsing under the requested descendant species succeeds.
    # - Descendant→ancestor conversion is never allowed.
    # - Results without .species: always pass.
    if species is not None and result is not None:
        expected = expected_species
        if expected is None:
            # species= was provided but doesn't resolve to a known species.
            # This is always an error — we can't validate against an unknown species.
            if raise_on_error:
                raise ParseError(f"Unknown species '{species}' passed to species= parameter")
            return None
        else:
            if type(result) is Species:
                matches_expected = result == expected
            elif hasattr(result, "species"):
                result_species = result.species
                if result_species == expected:
                    matches_expected = True
                elif result_species.is_ancestor_of(expected):
                    converted = _reparse_result_for_species(
                        parser=parser,
                        result=result,
                        expected_species=expected,
                        infer_class2_pairing=infer_class2_pairing,
                        required_result_types=required_result_types,
                        preferred_result_types=preferred_result_types,
                        only_class1=only_class1,
                        only_class2=only_class2,
                        max_allele_fields=max_allele_fields,
                    )
                    if converted is not None:
                        result = converted
                        matches_expected = True
                    else:
                        matches_expected = False
                else:
                    matches_expected = False
            else:
                matches_expected = True

            if not matches_expected:
                parsed_species = (
                    result.name
                    if type(result) is Species
                    else getattr(result.species, "name", None)
                )
                if raise_on_error:
                    raise ParseError(
                        f"Parsed species '{parsed_species}' does not match "
                        f"expected species '{expected.name}' for '{raw_string}'"
                    )
                return None

    return result



#!/usr/bin/env python3
"""Validate every allele name in cached official IMGT/IPD protein FASTA files."""

from __future__ import annotations

import argparse
import sys
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path

from mhcgnomes import Allele, parse

DEFAULT_FASTA_PATHS = (
    Path(".external-data/hla_prot.fasta"),
    Path(".external-data/MHC_prot.fasta"),
)


@dataclass(frozen=True)
class ValidationResult:
    path: Path
    record_count: int
    unique_name_count: int
    failures: tuple[str, ...]


def fasta_names(lines: Iterable[str]) -> Iterable[tuple[int, str]]:
    for line_number, line in enumerate(lines, start=1):
        if not line.startswith(">"):
            continue
        fields = line[1:].split()
        if len(fields) < 2:
            yield line_number, ""
        else:
            yield line_number, fields[1]


def validate_fasta(path: Path) -> ValidationResult:
    failures = []
    names = set()
    record_count = 0
    with path.open(encoding="utf-8") as handle:
        for line_number, name in fasta_names(handle):
            record_count += 1
            if not name:
                failures.append(f"line {line_number}: malformed FASTA header")
                continue
            names.add(name)
            try:
                result = parse(name, raise_on_error=False)
            except Exception as exc:  # report the complete upstream compatibility result
                failures.append(f"line {line_number}: {name!r} raised {exc!r}")
                continue
            if not isinstance(result, Allele):
                failures.append(f"line {line_number}: {name!r} parsed as {result!r}")
    if record_count == 0:
        failures.append("no FASTA headers found")
    return ValidationResult(path, record_count, len(names), tuple(failures))


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Parse every allele name in cached official IMGT/IPD protein FASTA files."
    )
    parser.add_argument("paths", nargs="*", type=Path, default=DEFAULT_FASTA_PATHS)
    parser.add_argument("--max-failures", type=int, default=20)
    return parser


def main(argv: Sequence[str] | None = None) -> int:
    args = build_parser().parse_args(list(argv) if argv is not None else None)
    failed = False
    for path in args.paths:
        if not path.is_file():
            print(
                f"ERROR: missing {path}; run 'mhcgnomes data download --group validation' first",
                file=sys.stderr,
            )
            failed = True
            continue
        result = validate_fasta(path)
        print(
            f"{path}: parsed {result.unique_name_count} unique allele names "
            f"from {result.record_count} FASTA records"
        )
        if result.failures:
            failed = True
            print(f"ERROR: {len(result.failures)} names failed:", file=sys.stderr)
            for failure in result.failures[: args.max_failures]:
                print(f"  {failure}", file=sys.stderr)
    return 1 if failed else 0


if __name__ == "__main__":
    raise SystemExit(main())



"""
Tests for how a species is chosen when the input string is ambiguous.

Two distinct sources of ambiguity are covered here:

1. A species prefix is inherited by every descendant, so a bare prefix matches
   an ancestor and everything under it ("BoLA" matches Bos sp. but also
   Bubalus bubalis). See https://github.com/pirl-unc/mhcgnomes/issues/103

2. A gene symbol with no species prefix can belong to many species, and the
   winner used to be the species with the most genes *visible* to it, which is
   inflated by whatever a broad parent group defines.
   See https://github.com/pirl-unc/mhcgnomes/issues/105
"""

import pytest

from mhcgnomes import MhcClass, Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# 1. A bare species prefix should not silently pick a descendant species
# ---------------------------------------------------------------------------

# (prefix, expected latin name) for every prefix which an ancestor shares with
# its descendants. Each of these used to resolve to an arbitrary descendant,
# chosen by the alphabetical order of the repr string.
PREFIX_TO_EXPECTED_SPECIES = [
    ("BoLA", "Bos sp."),
    ("CELA", "Cetacea sp."),
    ("ChLA", "Pan sp."),
    ("MusSp", "Mus sp."),
    # NHP resolves to Primata sp. because that is what the ontology maps the
    # prefix to. Worth noting that "NHP" means *non-human* primate while
    # Primata sp. is an ancestor of Homo sapiens; that mismatch predates this
    # change and is a curation question, not a parsing one.
    ("NHP", "Primata sp."),
    ("OmLA", "Aotus sp."),
    ("RT1", "Rattus sp."),
    ("RhLA", "Macaca sp."),
]


@pytest.mark.parametrize("prefix,expected", PREFIX_TO_EXPECTED_SPECIES)
@pytest.mark.parametrize("mhc_class", ["I", "II"])
def test_class_only_string_keeps_the_prefix_owner_species(prefix, expected, mhc_class):
    result = parse(f"{prefix} class {mhc_class}")
    eq_(type(result), MhcClass)
    eq_(result.mhc_class, mhc_class)
    eq_(result.species.name, expected)


@pytest.mark.parametrize("prefix,expected", PREFIX_TO_EXPECTED_SPECIES)
def test_class_only_string_agrees_with_bare_prefix(prefix, expected):
    """
    "<prefix> class I" and "<prefix>" describe the same species, so they must
    not disagree about which one it is.
    """
    # pin both sides, otherwise a curation change that moved them together
    # would keep this passing
    eq_(Species.get(prefix).name, expected)
    eq_(parse(f"{prefix} class I").species, Species.get(prefix))


def test_bola_class_i_is_cattle_not_water_buffalo():
    """
    BoLA is the *Bovine* Leukocyte Antigen system. Water buffalo is a separate
    genus which inherits the BoLA prefix, and it used to win this parse.
    """
    result = parse("BoLA class I")
    eq_(result.species.name, "Bos sp.")
    assert result.species.name != "Bubalus bubalis"


def test_bola_class_i_agrees_with_bola_allele():
    """The reported symptom: an allele said Bos, the class-only string said Bubalus."""
    eq_(parse("BoLA class I").species, parse("BoLA-N*01301").species)
    eq_(parse("BoLA class II").species, parse("BoLA-DRB3*011:01").species)


def test_explicit_descendant_prefix_still_resolves_to_that_descendant():
    """Preferring the ancestor must not make descendants unreachable."""
    eq_(parse("Bubu-DQA").species.name, "Bubalus bubalis")
    eq_(parse("Bota-DRB3*011:01").species.name, "Bos taurus")


# ---------------------------------------------------------------------------
# 2. Unprefixed gene symbols shared across species
# ---------------------------------------------------------------------------


def test_bare_BLB2_allele_resolves_to_chicken():
    """
    BLB1/BLB2 are the chicken MHC-B class II beta genes. Japanese quail only
    has them by inheritance from "Galliformes sp." -- its own ontology entry
    uses the Coja-DAB1/DBB1/DCB1 nomenclature -- but its larger inherited gene
    count used to win the tie.
    """
    eq_(parse("BLB2*02").species.name, "Gallus gallus")
    eq_(parse("BLB1*02").species.name, "Gallus gallus")


def test_bare_BLB2_agrees_with_explicit_chicken_prefix():
    eq_(parse("BLB2*02"), parse("Gaga-BLB2*02"))


def test_bare_bird_class1_and_class2_genes_agree_on_species():
    """
    Within one bird MHC region, bare BF2 inferred chicken while bare BLB2
    inferred quail. Both are chicken genes.
    """
    eq_(parse("BLB2*02").species, parse("BF2*02:01").species)


def test_quail_genes_still_resolve_to_quail():
    """The quail's own gene names must be unaffected."""
    eq_(parse("Coja-DAB1").species.name, "Coturnix japonica")
    eq_(parse("Coja-BLB2*02").species.name, "Coturnix japonica")


def test_declares_gene_distinguishes_own_genes_from_inherited_ones():
    quail = Species.get_by_latin_name("Coturnix japonica")
    chicken = Species.get_by_latin_name("Gallus gallus")
    buffalo = Species.get_by_latin_name("Bubalus bubalis")
    cattle = Species.get_by_latin_name("Bos sp.")

    # both birds can see BLB2, only chicken declares it
    assert "BLB2" in quail.gene_names
    assert "BLB2" in chicken.gene_names
    assert not quail.declares_gene("BLB2")
    assert chicken.declares_gene("BLB2")

    # and the quail's own class II beta nomenclature belongs to the quail
    assert quail.declares_gene("DBB1")
    assert not chicken.declares_gene("DBB1")

    # water buffalo can see the BoLA genes but declares none of them
    assert "NC1" in buffalo.gene_names
    assert not buffalo.declares_gene("NC1")
    assert cattle.declares_gene("NC1")


def test_declares_gene_is_case_normalizing_but_case_aware():
    """
    Gene lookup normalizes case, so "Ia1" (Paralichthys olivaceus) and "IA1"
    (Chrysolophus pictus) are the same key. Both species declare their own
    spelling; only one matches a given query exactly.
    """
    flounder = Species.get_by_latin_name("Paralichthys olivaceus")
    pheasant = Species.get_by_latin_name("Chrysolophus pictus")

    assert flounder.declares_gene("Ia1")
    assert pheasant.declares_gene("Ia1")
    assert flounder.declares_gene_with_same_case("Ia1")
    assert not pheasant.declares_gene_with_same_case("Ia1")
    assert pheasant.declares_gene_with_same_case("IA1")


def test_bare_gene_resolves_to_a_species_that_declares_it():
    """
    Every candidate under a broad parent group can see that group's genes, so
    the winner must be one that actually uses the name rather than whichever
    inheritor happens to have the largest gene list.
    """
    for gene_name, expected in [
        ("BLB2", "Gallus gallus"),
        ("BLB1", "Gallus gallus"),
        ("BF2", "Gallus gallus"),
        # quail's own class II beta genes stay with the quail
        ("DBB1", "Coturnix japonica"),
        # these were reassigned to an inheriting group when the ranking
        # measured gene-list size instead of declaration
        ("DAB1", "Crocodylus porosus"),
        ("Ia1", "Paralichthys olivaceus"),
    ]:
        result = parse(gene_name)
        eq_(result.species.name, expected)
        assert result.species.declares_gene(gene_name), gene_name


# ---------------------------------------------------------------------------
# Patr-AL is non-classical
# https://github.com/pirl-unc/mhcgnomes/issues/107
# ---------------------------------------------------------------------------


def test_patr_AL_is_non_classical():
    """
    Adams, Cooper & Parham (PMID 11564803) named AL a nonclassical class I
    molecule in the title of the paper describing it: three allotypes, present
    on ~50% of chimpanzee haplotypes, low expression. Being MHC-A-related is
    why it gets filed under A, but that does not make it classical.
    """
    eq_(parse("Patr-AL").mhc_class, "Ib")
    eq_(parse("ChLA-AL").mhc_class, "Ib")


def test_patr_A_is_still_classical():
    eq_(parse("Patr-A*01:01").mhc_class, "Ia")


def test_non_classical_A_related_loci_agree_across_primates():
    """AL should sit with the E/F/G family the ontology already calls Ib."""
    for name in ["HLA-E", "HLA-F", "HLA-G", "Mamu-E*02:11", "Caja-E", "Patr-AL"]:
        eq_(parse(name).mhc_class, "Ib")



import hashlib
import io
import json
import urllib.error
from pathlib import Path

import pytest

from mhcgnomes import external_data


def make_source(payload=b"source payload", **overrides):
    values = {
        "id": "example-source",
        "filename": "example.dat",
        "provider": "Example",
        "release": "1",
        "url": "https://official.example/example.dat",
        "sha256": hashlib.sha256(payload).hexdigest(),
        "license_url": "https://official.example/license",
        "groups": ("example-group",),
        "default": True,
    }
    values.update(overrides)
    return external_data.Source(**values)


def write_manifest(path: Path, sources):
    path.write_text(
        json.dumps(
            {
                "schema_version": 1,
                "sources": [
                    {
                        "id": source.id,
                        "filename": source.filename,
                        "provider": source.provider,
                        "release": source.release,
                        "url": source.url,
                        "sha256": source.sha256,
                        "license_url": source.license_url,
                        "groups": source.groups,
                        "default": source.default,
                    }
                    for source in sources
                ],
            }
        ),
        encoding="utf-8",
    )


def test_shipped_manifest_is_valid_and_uses_official_providers():
    sources = external_data.load_manifest()
    assert len(sources) == 7
    assert {source.provider for source in sources} == {"IPD-IMGT/HLA", "IPD-MHC"}
    assert all("/ANHIG/" in source.url for source in sources)


@pytest.mark.parametrize("arguments", [[], ["list"]])
def test_data_command_lists_sources_by_default_without_downloading(arguments, monkeypatch, capsys):
    monkeypatch.setattr(
        external_data,
        "_download_to_cache",
        lambda *_args, **_kwargs: pytest.fail("listing must not download data"),
    )

    assert external_data.main(arguments) == 0

    output = capsys.readouterr().out
    assert "imgt-hla-allele-history-3.42.0" in output
    assert "ipd-mhc-3.8.0.0-protein" in output


def test_select_sources_defaults_and_unions_explicit_groups():
    default = make_source()
    optional = make_source(
        id="optional", filename="optional.dat", groups=("optional-group",), default=False
    )
    sources = (default, optional)

    assert external_data.select_sources(sources) == (default,)
    assert external_data.select_sources(sources, groups=("optional-group",)) == (optional,)
    assert (
        external_data.select_sources(sources, source_ids=(default.id,), groups=("optional-group",))
        == sources
    )
    assert external_data.select_sources(sources, all_sources=True) == sources


def test_select_sources_rejects_unknown_values():
    source = make_source()
    with pytest.raises(external_data.ExternalDataError, match="Unknown source IDs"):
        external_data.select_sources((source,), source_ids=("missing",))
    with pytest.raises(external_data.ExternalDataError, match="Unknown source groups"):
        external_data.select_sources((source,), groups=("missing",))


def test_verified_destination_avoids_cache_and_network(tmp_path, monkeypatch):
    payload = b"already materialized"
    source = make_source(payload)
    destination = tmp_path / "destination"
    destination.mkdir()
    (destination / source.filename).write_bytes(payload)

    monkeypatch.setattr(
        external_data,
        "_download_to_cache",
        lambda *_args, **_kwargs: pytest.fail("network should not be used"),
    )
    path, origin = external_data.resolve_source(
        source, destination_dir=destination, cache_dir=tmp_path / "cache"
    )

    assert path.read_bytes() == payload
    assert origin == "destination"


def test_content_cache_is_materialized_without_network(tmp_path, monkeypatch):
    payload = b"cached content"
    source = make_source(payload)
    cache = tmp_path / "cache"
    cache.mkdir()
    (cache / source.sha256).write_bytes(payload)
    monkeypatch.setattr(
        external_data,
        "_download_to_cache",
        lambda *_args, **_kwargs: pytest.fail("network should not be used"),
    )

    path, origin = external_data.resolve_source(
        source, destination_dir=tmp_path / "destination", cache_dir=cache
    )

    assert path.read_bytes() == payload
    assert origin == "cache"


def test_local_mirror_accepts_filename_or_content_hash(tmp_path):
    payload = b"mirrored content"
    source = make_source(payload)
    mirror = tmp_path / "mirror"
    mirror.mkdir()
    (mirror / source.sha256).write_bytes(payload)

    path, origin = external_data.resolve_source(
        source,
        destination_dir=tmp_path / "destination",
        cache_dir=tmp_path / "cache",
        mirrors=(str(mirror),),
        offline=True,
    )

    assert path.read_bytes() == payload
    assert origin == f"mirror {mirror}"


def test_http_mirror_precedes_official_source(tmp_path, monkeypatch):
    payload = b"HTTP mirror content"
    source = make_source(payload)
    requested_urls = []

    def fake_download(url, _source, cache_path, retries=3):
        requested_urls.append(url)
        cache_path.parent.mkdir(parents=True, exist_ok=True)
        cache_path.write_bytes(payload)

    monkeypatch.setattr(external_data, "_download_to_cache", fake_download)
    path, origin = external_data.resolve_source(
        source,
        destination_dir=tmp_path / "destination",
        cache_dir=tmp_path / "cache",
        mirrors=("https://cache.example/base",),
    )

    assert path.read_bytes() == payload
    assert requested_urls == ["https://cache.example/base/example.dat"]
    assert origin == "mirror https://cache.example/base"


def test_http_mirror_can_expose_content_by_hash(tmp_path, monkeypatch):
    payload = b"content-addressed HTTP mirror"
    source = make_source(payload)
    requested_urls = []

    def fake_download(url, _source, cache_path, retries=3):
        requested_urls.append(url)
        if not url.endswith(source.sha256):
            raise external_data.ExternalDataError("filename not present")
        cache_path.parent.mkdir(parents=True, exist_ok=True)
        cache_path.write_bytes(payload)

    monkeypatch.setattr(external_data, "_download_to_cache", fake_download)
    path, origin = external_data.resolve_source(
        source,
        destination_dir=tmp_path / "destination",
        cache_dir=tmp_path / "cache",
        mirrors=("https://cache.example/base",),
    )

    assert path.read_bytes() == payload
    assert requested_urls == [
        "https://cache.example/base/example.dat",
        f"https://cache.example/base/{source.sha256}",
    ]
    assert origin == "mirror https://cache.example/base"


def test_offline_mode_skips_network_mirror_and_primary(tmp_path, monkeypatch):
    source = make_source()
    monkeypatch.setattr(
        external_data,
        "_download_to_cache",
        lambda *_args, **_kwargs: pytest.fail("network should not be used"),
    )
    with pytest.raises(external_data.ExternalDataError, match="Offline"):
        external_data.resolve_source(
            source,
            destination_dir=tmp_path / "destination",
            cache_dir=tmp_path / "cache",
            mirrors=("https://cache.example",),
            offline=True,
        )


def test_primary_download_is_verified_and_reused(tmp_path, monkeypatch):
    payload = b"official content"
    source = make_source(payload)
    requests = []

    def fake_urlopen(request, timeout):
        requests.append((request.full_url, request.headers["User-agent"], timeout))
        return io.BytesIO(payload)

    monkeypatch.setattr(external_data.urllib.request, "urlopen", fake_urlopen)
    cache = tmp_path / "cache"
    first, origin = external_data.resolve_source(
        source, destination_dir=tmp_path / "first", cache_dir=cache
    )
    second, second_origin = external_data.resolve_source(
        source, destination_dir=tmp_path / "second", cache_dir=cache
    )

    assert first.read_bytes() == payload
    assert second.read_bytes() == payload
    assert origin == "official source"
    assert second_origin == "cache"
    assert requests == [(source.url, external_data.USER_AGENT, 60)]


def test_checksum_mismatch_is_rejected_without_poisoning_cache(tmp_path, monkeypatch):
    source = make_source(b"expected")
    monkeypatch.setattr(
        external_data.urllib.request,
        "urlopen",
        lambda *_args, **_kwargs: io.BytesIO(b"tampered"),
    )
    cache_path = tmp_path / "cache" / source.sha256

    with pytest.raises(external_data.ExternalDataError, match="Checksum mismatch"):
        external_data._download_to_cache(source.url, source, cache_path)

    assert not cache_path.exists()


def test_nontransient_http_error_is_not_retried(tmp_path, monkeypatch):
    source = make_source()
    requests = []

    def not_found(request, timeout):
        requests.append((request.full_url, timeout))
        raise urllib.error.HTTPError(request.full_url, 404, "not found", {}, None)

    monkeypatch.setattr(external_data.urllib.request, "urlopen", not_found)
    monkeypatch.setattr(
        external_data.time,
        "sleep",
        lambda *_args: pytest.fail("a permanent HTTP error must not be retried"),
    )

    with pytest.raises(external_data.ExternalDataError, match="Failed to download"):
        external_data._download_to_cache(source.url, source, tmp_path / "cache" / source.sha256)

    assert requests == [(source.url, 60)]


def test_cli_can_materialize_from_local_mirror_offline(tmp_path, capsys):
    payload = b"CLI payload"
    source = make_source(payload)
    manifest = tmp_path / "manifest.json"
    mirror = tmp_path / "mirror"
    mirror.mkdir()
    (mirror / source.filename).write_bytes(payload)
    write_manifest(manifest, (source,))
    destination = tmp_path / "destination"

    result = external_data.main(
        [
            "download",
            "--manifest",
            str(manifest),
            "--cache-dir",
            str(tmp_path / "cache"),
            "--destination",
            str(destination),
            "--mirror",
            str(mirror),
            "--offline",
        ]
    )

    assert result == 0
    assert (destination / source.filename).read_bytes() == payload
    assert "mirror" in capsys.readouterr().out



#!/usr/bin/env python3
"""
Generate serotype-to-allele mappings from the IPD-IMGT/HLA dictionary.

Source: https://www.ebi.ac.uk/ipd/imgt/hla/alleles/dictionary/
The HLA dictionary contains WHO-assigned serotypes for HLA alleles.

This script parses the Excel file and generates YAML mappings.
"""

import re
from collections import defaultdict
from pathlib import Path

import pandas as pd
import yaml


def parse_serotype(serotype_str: str) -> list[str]:
    """
    Parse a serotype string, handling complex formats.

    Examples:
        "A1" -> ["A1"]
        "A23(9)" -> ["A23"]  # Split of A9
        "A2/19" -> ["A2", "A19"]  # Multiple assignments
        "Low A24(9)" -> ["A24"]  # Ignore "Low" prefix
        "-" -> []  # No assignment
        "Null" -> []  # Null allele
        "Not expressed" -> []
        "?B44(12)" -> []  # Skip uncertain assignments
        "Soluble" -> []  # Skip soluble designation
        "Cw9(w3)" -> ["Cw9"]  # Handle workshop split notation
    """
    if not serotype_str or serotype_str in ["-", "Null", "Not expressed", "Soluble"]:
        return []

    # Skip uncertain assignments (starting with ?)
    if serotype_str.startswith("?"):
        return []

    # Remove "Low " prefix
    serotype_str = re.sub(r"^Low\s+", "", serotype_str)

    # Handle split serotypes like "A23(9)" or "Cw9(w3)" -> extract base serotype
    # The content in parens is the broad serotype it splits from
    serotype_str = re.sub(r"\([wW]?\d+\)", "", serotype_str)

    # Handle percentage confidence like "A1 [98-100]" -> "A1"
    serotype_str = re.sub(r"\s*\[[\d\-]+\]", "", serotype_str)

    # Split on "/" for multiple assignments
    parts = serotype_str.split("/")

    serotypes = []
    for part in parts:
        part = part.strip()
        # Skip if it's just a number (like in "A2/19" where 19 means A19)
        if part.isdigit():
            continue
        # Skip empty or "blank"
        if not part or part.lower() == "blank":
            continue
        serotypes.append(part)

    return serotypes


def normalize_allele_name(allele: str) -> str:
    """
    Normalize allele name to two-field format for consistency.

    Examples:
        "A*01:01:01:01" -> "A*0101"
        "A*01:01:02N" -> "A*0101"  # Remove suffix
        "DRB1*01:01" -> "DRB1*0101"
        "B*15:102" -> "B*15102"  # Three-digit second field (kept as-is)
    """
    # Remove any suffix (N for null, L for low, etc.)
    allele = re.sub(r"[NLSQA]$", "", allele)

    # Parse the allele format
    match = re.match(r"([A-Z0-9]+)\*(\d+):(\d+)", allele)
    if match:
        gene = match.group(1)
        field1 = match.group(2).zfill(2)
        field2 = match.group(3)
        # Keep 3-digit second fields as-is (e.g., 102, 103)
        # For 1-2 digit fields, pad to 2 digits
        if len(field2) <= 2:
            field2 = field2.zfill(2)
        return f"{gene}*{field1}{field2}"

    return allele


def load_hla_dictionary(xlsx_path: Path) -> pd.DataFrame:
    """Load the HLA dictionary Excel file."""
    return pd.read_excel(xlsx_path, sheet_name="A")


def build_serotype_mappings(df: pd.DataFrame) -> dict[str, dict[str, list[str]]]:
    """
    Build serotype-to-allele mappings from the HLA dictionary.

    Returns a nested dict: {gene: {serotype: [alleles]}}
    """
    # Group alleles by their WHO-assigned serotype
    serotype_to_alleles: dict[str, set[str]] = defaultdict(set)

    for _, row in df.iterrows():
        allele = row["HLA Allele"]
        who_serotype = row.get("WHO Assigned Type", "")

        # Skip if no WHO assignment
        if pd.isna(who_serotype):
            continue

        serotypes = parse_serotype(str(who_serotype))
        if not serotypes:
            continue

        # Normalize the allele name
        normalized = normalize_allele_name(allele)

        for serotype in serotypes:
            serotype_to_alleles[serotype].add(normalized)

    # Organize by gene prefix (A, B, C, DR, DQ, DP)
    result: dict[str, dict[str, list[str]]] = {}

    for serotype, alleles in sorted(serotype_to_alleles.items()):
        # Sort alleles for consistent output
        sorted_alleles = sorted(alleles)
        result[serotype] = sorted_alleles

    return result


def generate_yaml(mappings: dict[str, list[str]], existing_yaml_path: Path) -> str:
    """Generate YAML output, preserving non-HLA entries and manually curated HLA entries."""
    # Load existing YAML to preserve non-HLA entries and manual HLA curation
    with open(existing_yaml_path) as f:
        existing = yaml.safe_load(f)

    # Extract non-HLA entries
    non_hla = {k: v for k, v in existing.items() if k != "HLA"}

    # Build HLA section
    hla_section = {}

    # Add special entries first (Bw4, Bw6, Aw68, Aw69)
    # These are public epitopes/workshop antigens that we'll preserve manually
    special_serotypes = ["Aw68", "Aw69", "Bw4", "Bw6"]
    if "HLA" in existing:
        for special in special_serotypes:
            if special in existing["HLA"]:
                hla_section[special] = existing["HLA"][special]

    # Add WHO-assigned serotypes from dictionary
    for serotype, alleles in sorted(mappings.items()):
        if serotype not in special_serotypes:
            hla_section[serotype] = alleles

    # Preserve existing HLA entries that aren't in the dictionary
    # (e.g., DP workshop serotypes, some C serotypes)
    if "HLA" in existing:
        for serotype, alleles in existing["HLA"].items():
            if serotype not in hla_section:
                hla_section[serotype] = alleles

    # Build output
    output = {"HLA": hla_section}
    output.update(non_hla)

    return output


def main():
    script_dir = Path(__file__).parent
    data_dir = script_dir.parent / "mhcgnomes" / "data"
    xlsx_path = data_dir / "hla_dictionary.xlsx"
    yaml_path = data_dir / "serotypes.yaml"

    print(f"Loading HLA dictionary from {xlsx_path}")
    df = load_hla_dictionary(xlsx_path)
    print(f"Loaded {len(df)} alleles")

    print("Building serotype mappings...")
    mappings = build_serotype_mappings(df)
    print(f"Found {len(mappings)} serotypes")

    for serotype, alleles in sorted(mappings.items())[:10]:
        print(f"  {serotype}: {len(alleles)} alleles")
    print("  ...")

    print(f"Generating YAML, preserving non-HLA from {yaml_path}")
    output = generate_yaml(mappings, yaml_path)

    # Write to a new file for review
    output_path = data_dir / "serotypes_generated.yaml"

    # Custom YAML dump for better formatting
    with open(output_path, "w") as f:
        f.write(
            """############################################################
#
# Serotypes and the list of alleles they contain.
# Generated from IPD-IMGT/HLA Dictionary:
# https://www.ebi.ac.uk/ipd/imgt/hla/alleles/dictionary/
#
# The WHO Assigned Type column is used for serotype assignments.
#
############################################################

"""
        )

        # Write HLA section
        f.write("HLA:\n")
        hla = output["HLA"]

        # Write special serotypes first with comments
        special_with_comments = {
            "Aw68": "  # Old workshop serotype (now A68)",
            "Aw69": "  # Old workshop serotype (now A69)",
            "Bw4": "  # Public epitope (supertype)",
            "Bw6": "  # Public epitope (supertype)",
        }

        for special, comment in special_with_comments.items():
            if special in hla:
                f.write(f"  {special}:{comment}\n")
                for allele in hla[special]:
                    f.write(f"    - {allele}\n")
                f.write("\n")

        # Write remaining serotypes
        for serotype in sorted(hla.keys()):
            if serotype in special_with_comments:
                continue
            alleles = hla[serotype]
            f.write(f"  {serotype}:\n")
            for allele in alleles:
                f.write(f"    - {allele}\n")

        # Write non-HLA sections
        for species, serotypes in output.items():
            if species == "HLA":
                continue
            f.write(f"\n{species}:\n")
            for serotype, alleles in serotypes.items():
                if alleles:
                    f.write(f"  {serotype}:\n")
                    for allele in alleles:
                        f.write(f"    - {allele}\n")
                else:
                    f.write(f"  {serotype}: []\n")

    print(f"Written to {output_path}")
    print("\nReview the generated file, then copy to serotypes.yaml if correct.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Collect MHC allele strings from open-access papers, targeting species diversity.

v2: Focuses on PMC open access, adds species-targeted queries, and
generates a manual review set for each paper.

Usage:
    python paper/scripts/collect_all_v2.py
"""

import csv
import os
import re
import subprocess
import sys
import time
import urllib.request
import urllib.error
from pathlib import Path

try:
    from Bio import Entrez
except ImportError:
    print("Requires biopython: pip install biopython", file=sys.stderr)
    sys.exit(1)

Entrez.email = "mhcgnomes-paper@example.com"

ROOT = Path(__file__).resolve().parent.parent.parent
RAW_DIR = ROOT / "paper" / "raw"
VAL_DIR = ROOT / "paper" / "validation"
REVIEW_DIR = ROOT / "paper" / "review"
SCRAPE_SCRIPT = ROOT / "paper" / "scripts" / "scrape_paper.py"

# Open-access focused queries targeting species diversity.
# "open access"[filter] restricts to PMC OA subset.
QUERIES = {
    # ── Birds ──
    "bird_passerine": '"MHC class II" AND (passerine OR warbler OR finch OR tit OR sparrow) AND allele AND 2018:2025[dp]',
    "bird_raptor": '"MHC" AND (raptor OR eagle OR falcon OR owl OR hawk) AND (diversity OR allele) AND 2018:2025[dp]',
    "bird_seabird": '"MHC" AND (penguin OR petrel OR albatross OR gull OR tern) AND (diversity OR allele) AND 2018:2025[dp]',
    "bird_galliform": '("MHC" OR "BLB" OR "BF") AND (chicken OR turkey OR quail OR pheasant OR grouse) AND (allele OR haplotype) AND 2018:2025[dp]',
    "bird_waterfowl": '"MHC" AND (duck OR goose OR swan) AND (diversity OR allele) AND 2018:2025[dp]',
    # ── Fish ──
    "fish_salmonid": '("MHC" OR "UBA" OR "DAB") AND (salmon OR trout OR salmonid) AND (allele OR genotyping) AND 2018:2025[dp]',
    "fish_cichlid": '"MHC" AND cichlid AND (diversity OR allele) AND 2018:2025[dp]',
    "fish_other": '"MHC" AND (carp OR zebrafish OR tilapia OR cod OR stickleback OR catfish) AND (allele OR diversity) AND 2018:2025[dp]',
    # ── Reptiles & amphibians ──
    "reptile_turtle": '"MHC" AND (turtle OR tortoise) AND (diversity OR allele) AND 2015:2025[dp]',
    "reptile_lizard_snake": '"MHC" AND (lizard OR snake OR iguana OR gecko) AND (diversity OR allele) AND 2015:2025[dp]',
    "reptile_croc": '"MHC" AND (crocodile OR alligator OR caiman) AND 2015:2025[dp]',
    "amphibian": '"MHC" AND (frog OR toad OR salamander OR newt OR Xenopus) AND (diversity OR allele) AND 2018:2025[dp]',
    # ── Mammals (non-model) ──
    "mammal_bat": '"MHC" AND (bat OR chiroptera OR Myotis) AND (diversity OR allele) AND 2018:2025[dp]',
    "mammal_marsupial": '"MHC" AND (marsupial OR koala OR possum OR wallaby OR devil) AND (diversity OR allele) AND 2015:2025[dp]',
    "mammal_marine": '"MHC" AND (whale OR dolphin OR seal OR pinniped) AND (diversity OR allele) AND 2018:2025[dp]',
    "mammal_carnivore": '("MHC" OR "DLA" OR "FLA") AND (wolf OR cheetah OR leopard OR bear) AND (diversity OR allele) AND 2018:2025[dp]',
    "mammal_ungulate": '("MHC" OR "BoLA" OR "SLA" OR "OLA" OR "ELA") AND (cattle OR pig OR sheep OR horse OR deer) AND (genotyping OR allele) AND 2018:2025[dp]',
    "mammal_primate": '("MHC" OR "Mamu" OR "Mafa") AND (macaque OR chimpanzee OR gorilla OR marmoset) AND (genotyping OR allele) AND 2018:2025[dp]',
    "mammal_rodent": '("MHC" OR "H-2") AND (hamster OR vole OR squirrel OR guinea) AND diversity AND 2018:2025[dp]',
    # ── Conservation ──
    "conservation": '"MHC diversity" AND (endangered OR conservation OR bottleneck) AND allele AND 2018:2025[dp]',
    # ── Human (population) ──
    "human_population": '"HLA" AND "allele frequency" AND population AND genotyping AND 2020:2025[dp]',
    # ── Shark ──
    "shark": '"MHC" AND (shark OR elasmobranch) AND 2010:2025[dp]',
}



def search_pubmed(query, max_results=15):
    try:
        handle = Entrez.esearch(
            db="pmc", term=query, retmax=max_results, sort="relevance"
        )
        result = Entrez.read(handle)
        handle.close()
        return result["IdList"]
    except Exception as e:
        print(f"  Search failed: {e}", file=sys.stderr)
        return []


def fetch_paper_metadata(pmc_ids):
    """Fetch metadata for PMC IDs. Returns basic info from PMC summaries."""
    if not pmc_ids:
        return []
    papers = []
    try:
        handle = Entrez.esummary(db="pmc", id=",".join(pmc_ids))
        records = Entrez.read(handle)
        handle.close()
        for rec in records:
            papers.append({
                "pmid": str(rec.get("Id", "")),
                "first_author": str(rec.get("AuthorList", [""])[0]) if rec.get("AuthorList") else "",
                "year": str(rec.get("PubDate", ""))[:4],
                "title": str(rec.get("Title", "")),
                "journal": str(rec.get("FullJournalName", rec.get("Source", ""))),
                "doi": str(rec.get("DOI", "")),
            })
    except Exception as e:
        print(f"  Metadata fetch failed: {e}", file=sys.stderr)
        # Fall back to just using IDs
        for pmc_id in pmc_ids:
            papers.append({
                "pmid": pmc_id, "first_author": "", "year": "",
                "title": "", "journal": "", "doi": "",
            })
    return papers


def pmid_to_pmc(pmid):
    try:
        handle = Entrez.elink(dbfrom="pubmed", db="pmc", id=pmid)
        result = Entrez.read(handle)
        handle.close()
        for linkset in result:
            for db_link in linkset.get("LinkSetDb", []):
                if db_link.get("DbTo") == "pmc":
                    links = db_link.get("Link", [])
                    if links:
                        return links[0]["Id"]
    except Exception:
        pass
    return None


def download_pmc_package(pmc_id, raw_dir):
    """Download PMC OA package and extract supplementary files.

    Returns list of extracted file paths.
    """
    import tarfile

    # Get FTP URL from OA API
    try:
        oa_url = f"https://www.ncbi.nlm.nih.gov/pmc/utils/oa/oa.fcgi?id=PMC{pmc_id}"
        req = urllib.request.Request(oa_url, headers={"User-Agent": "mhcgnomes-paper/1.0"})
        with urllib.request.urlopen(req, timeout=15) as resp:
            text = resp.read().decode()
    except Exception:
        return []

    # Extract FTP URL
    m = re.search(r'href="(ftp://[^"]+\.tar\.gz)"', text)
    if not m:
        # Try HTTPS fallback
        m = re.search(r'href="(https?://[^"]+\.tar\.gz)"', text)
    if not m:
        return []

    pkg_url = m.group(1)
    # Convert ftp to https for reliability
    pkg_url = pkg_url.replace("ftp://ftp.ncbi.nlm.nih.gov/", "https://ftp.ncbi.nlm.nih.gov/")

    # Download the package
    pkg_path = raw_dir / f"PMC{pmc_id}.tar.gz"
    if not pkg_path.exists():
        try:
            req = urllib.request.Request(pkg_url, headers={"User-Agent": "mhcgnomes-paper/1.0"})
            with urllib.request.urlopen(req, timeout=60) as resp:
                with open(pkg_path, "wb") as f:
                    f.write(resp.read())
        except Exception as e:
            print(f"    Download failed: {e}", file=sys.stderr)
            return []

    # Extract supplementary files
    extracted = []
    try:
        with tarfile.open(pkg_path, "r:gz") as tar:
            for member in tar.getmembers():
                name_lower = member.name.lower()
                if any(name_lower.endswith(ext) for ext in
                       [".xlsx", ".xls", ".csv", ".tsv", ".txt", ".docx"]):
                    # Extract to raw_dir
                    member.name = Path(member.name).name  # flatten path
                    tar.extract(member, raw_dir)
                    extracted.append(raw_dir / member.name)
    except Exception as e:
        print(f"    Extract failed: {e}", file=sys.stderr)

    return extracted


def download_file(url, dest):
    try:
        req = urllib.request.Request(url, headers={"User-Agent": "mhcgnomes-paper/1.0"})
        with urllib.request.urlopen(req, timeout=30) as resp:
            with open(dest, "wb") as f:
                f.write(resp.read())
        return True
    except Exception as e:
        print(f"    Download failed: {e}", file=sys.stderr)
        return False


def scrape_file(filepath, species, source, output):
    try:
        result = subprocess.run(
            [
                sys.executable, str(SCRAPE_SCRIPT),
                "--input", str(filepath),
                "--species", species,
                "--source", source,
                "--output", str(output),
            ],
            capture_output=True, text=True, timeout=60,
        )
        if result.returncode == 0:
            return sum(1 for _ in open(output)) - 1
    except Exception:
        pass
    return 0


def generate_review_file(scrape_tsv, review_path, pmid, title):
    """Create a review TSV with parsed/unparsed status for manual audit."""
    try:
        from mhcgnomes import parse as mhc_parse
    except ImportError:
        return

    rows = []
    with open(scrape_tsv) as f:
        reader = csv.DictReader(f, delimiter="\t")
        for row in reader:
            raw = row["raw_string"]
            result = mhc_parse(raw, raise_on_error=False)
            rows.append({
                "raw_string": raw,
                "parsed": "yes" if result else "no",
                "parsed_as": str(result) if result else "",
                "result_type": type(result).__name__ if result else "",
                "species": result.species.name if result and hasattr(result, "species") and result.species else "",
                "correct": "",  # for manual review
                "notes": "",    # for manual review
                "source": f"PMC:{pmid}",
            })

    with open(review_path, "w") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=["raw_string", "parsed", "parsed_as", "result_type",
                         "species", "correct", "notes", "source"],
            delimiter="\t",
        )
        writer.writeheader()
        writer.writerows(rows)


def main():
    REVIEW_DIR.mkdir(parents=True, exist_ok=True)

    # Step 1: Run all queries
    all_pmids = set()
    stratum_papers = {}

    for stratum, query in QUERIES.items():
        print(f"Searching: {stratum}", file=sys.stderr)
        pmids = search_pubmed(query, max_results=15)
        print(f"  Found {len(pmids)} papers", file=sys.stderr)
        new_pmids = [p for p in pmids if p not in all_pmids]
        all_pmids.update(pmids)

        if new_pmids:
            papers = fetch_paper_metadata(new_pmids)
            for p in papers:
                p["stratum"] = stratum
            stratum_papers[stratum] = papers
        time.sleep(0.4)

    # Save expanded candidate list
    all_papers = []
    for papers in stratum_papers.values():
        all_papers.extend(papers)

    candidates_path = VAL_DIR / "candidate_papers_v2.tsv"
    with open(candidates_path, "w") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=["stratum", "pmid", "first_author", "year",
                         "title", "journal", "doi"],
            delimiter="\t",
        )
        writer.writeheader()
        writer.writerows(all_papers)
    print(f"\n{len(all_papers)} candidate papers → {candidates_path}\n", file=sys.stderr)

    # Step 2: Process each paper
    total_strings = 0
    papers_with_data = 0
    consolidated_rows = []

    for i, paper in enumerate(all_papers):
        pmid = paper["pmid"]
        stratum = paper["stratum"]
        title = paper.get("title", "")[:60]

        print(f"[{i+1}/{len(all_papers)}] PMC{pmid} ({stratum}) {title}...", file=sys.stderr)

        safe_id = f"PMC_{pmid}"
        raw_dir = RAW_DIR / safe_id
        raw_dir.mkdir(parents=True, exist_ok=True)

        pmc_id = pmid  # Already a PMC ID from PMC search

        # Download OA package and extract supplementary files
        extracted = download_pmc_package(pmc_id, raw_dir)
        if not extracted:
            time.sleep(0.3)
            continue

        print(f"  {len(extracted)} supplementary files extracted", file=sys.stderr)

        paper_strings = 0
        paper_tsvs = []

        for dest in extracted:
            out_tsv = VAL_DIR / f"{safe_id}_{dest.stem}.tsv"
            n = scrape_file(dest, "", f"PMC:{pmid}", out_tsv)
            if n > 0:
                print(f"  {n} strings from {dest.name}", file=sys.stderr)
                paper_strings += n
                paper_tsvs.append(out_tsv)
                with open(out_tsv) as f:
                    reader = csv.DictReader(f, delimiter="\t")
                    for row in reader:
                        consolidated_rows.append(row)

        if paper_strings > 0:
            papers_with_data += 1
            total_strings += paper_strings

            # Generate review file for this paper
            # Merge all TSVs for this paper into one review file
            review_path = REVIEW_DIR / f"{safe_id}_review.tsv"
            merged_tsv = VAL_DIR / f"{safe_id}_merged.tsv"
            with open(merged_tsv, "w") as f:
                writer = csv.DictWriter(
                    f, fieldnames=["raw_string", "expected_species", "source"],
                    delimiter="\t",
                )
                writer.writeheader()
                for tsv in paper_tsvs:
                    with open(tsv) as tf:
                        reader = csv.DictReader(tf, delimiter="\t")
                        writer.writerows(reader)

            generate_review_file(merged_tsv, review_path, f"PMC{pmid}", paper["title"])
            review_lines = sum(1 for _ in open(review_path)) - 1
            print(f"  Review file: {review_path} ({review_lines} entries)", file=sys.stderr)

        time.sleep(0.5)

    # Write consolidated file
    consolidated = VAL_DIR / "all_papers_v2.tsv"
    with open(consolidated, "w") as f:
        writer = csv.DictWriter(
            f, fieldnames=["raw_string", "expected_species", "source"],
            delimiter="\t",
        )
        writer.writeheader()
        writer.writerows(consolidated_rows)

    print(f"\n{'='*60}", file=sys.stderr)
    print(f"Papers with data:  {papers_with_data}", file=sys.stderr)
    print(f"Total strings:     {total_strings}", file=sys.stderr)
    print(f"Consolidated:      {consolidated}", file=sys.stderr)
    print(f"Review files:      {REVIEW_DIR}/", file=sys.stderr)


if __name__ == "__main__":
    main()



"""
Tests for how a caller can tell a explicit species from an inferred one, and for
comparing two species names that differ in specificity.

https://github.com/pirl-unc/mhcgnomes/issues/116
https://github.com/pirl-unc/mhcgnomes/issues/118
"""

import pytest

from mhcgnomes import Allele, Gene, Pair, ParseError, Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# 1. species_source
# ---------------------------------------------------------------------------

EXPLICIT = [
    "HLA-A*02:01",  # attached prefix
    "Gaga-BLB2*02",
    "Coja-DBB1",
    "BoLA-N*01301",
    "Bota-DRB3*011:01",
    "H2-Kb",
    "BoLA class I",
    "mouse H2-Kb",  # leading common name
    "Homo sapiens class I",  # leading latin name
    "HLA",  # species alone
]

INFERRED = [
    "BLB2*02",  # species comes from the gene name
    "BF2*02:01",
    "B12 class I",
]

DEFAULTED = [
    "A*02:01",  # no species anywhere, falls back to default_species
    "MHC class II",
    "class II",
    "TAP1",
]


@pytest.mark.parametrize("name", EXPLICIT)
def test_explicit_species_is_reported_as_explicit(name):
    eq_(parse(name).species_source, "explicit")


@pytest.mark.parametrize("name", INFERRED)
def test_species_from_a_gene_name_is_reported_as_inferred(name):
    eq_(parse(name).species_source, "inferred")


@pytest.mark.parametrize("name", DEFAULTED)
def test_species_from_default_species_is_reported_as_default(name):
    eq_(parse(name).species_source, "default")


@pytest.mark.parametrize("name", EXPLICIT)
def test_explicit_species_is_from_input(name):
    assert parse(name).species_from_input


@pytest.mark.parametrize("name", INFERRED + DEFAULTED)
def test_non_explicit_species_is_not_from_input(name):
    assert not parse(name).species_from_input


@pytest.mark.parametrize("name", EXPLICIT + INFERRED + DEFAULTED)
def test_species_from_input_agrees_with_species_source(name):
    result = parse(name)
    eq_(result.species_from_input, result.species_source == "explicit")


def test_default_is_distinguishable_from_inferred():
    """
    The two are different problems: 'default' means no species was involved at
    all, 'inferred' means one was derived from a gene name. A caller putting
    human MHC on a fish study needs to tell them apart.
    """
    eq_(parse("MHC class II").species_source, "default")
    eq_(parse("BLB2*02").species_source, "inferred")


def test_default_species_argument_is_respected():
    result = parse("A*02:01", default_species="Mamu")
    eq_(result.species_source, "default")
    eq_(result.species.name, "Macaca mulatta")


# ---------------------------------------------------------------------------
# 2. Provenance must not leak into identity
# ---------------------------------------------------------------------------


def test_provenance_does_not_affect_equality_or_hashing():
    explicit = parse("HLA-A*02:01")
    defaulted = parse("A*02:01")
    eq_(explicit.species_source, "explicit")
    eq_(defaulted.species_source, "default")
    eq_(explicit, defaulted)
    eq_(hash(explicit), hash(defaulted))


def test_provenance_does_not_appear_in_repr_or_dict():
    result = parse("HLA-A*02:01")
    assert "species_source" not in repr(result)
    assert "species_source" not in result.to_dict()


def test_directly_constructed_results_have_no_provenance():
    assert Allele.get("HLA", "A", "02", "01").species_source is None
    assert Gene.get("HLA", "A").species_source is None


# ---------------------------------------------------------------------------
# 3. require_explicit_species
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("name", EXPLICIT)
def test_require_explicit_species_keeps_explicit_results(name):
    eq_(parse(name, require_explicit_species=True), parse(name))


@pytest.mark.parametrize("name", INFERRED + DEFAULTED)
def test_require_explicit_species_rejects_non_explicit_results(name):
    assert parse(name, require_explicit_species=True, raise_on_error=False) is None


@pytest.mark.parametrize("name", INFERRED + DEFAULTED)
def test_require_explicit_species_raises_when_asked(name):
    with pytest.raises(ParseError):
        parse(name, require_explicit_species=True)


def test_species_source_is_none_for_unparsed_input():
    """
    Every parseable result currently carries a species, so the None case only
    arises for input that does not parse at all.
    """
    assert parse("n/a", raise_on_error=False) is None
    assert parse("n/a", require_explicit_species=True, raise_on_error=False) is None


def test_require_explicit_species_composes_with_required_result_types():
    eq_(
        parse(
            "Gaga-BLB2*02",
            required_result_types=(Allele, Gene, Pair),
            require_explicit_species=True,
        ).to_string(),
        "Gaga-BLB2*02",
    )
    assert (
        parse(
            "BLB2*02",
            required_result_types=(Allele, Gene, Pair),
            require_explicit_species=True,
            raise_on_error=False,
        )
        is None
    )


# ---------------------------------------------------------------------------
# 4. Species.compatible_with
# ---------------------------------------------------------------------------

COMPATIBLE = [
    ("Bos taurus", "Bos sp."),
    ("Sus scrofa", "Sus sp."),
    ("Coturnix japonica", "Galliformes sp."),
    ("Homo sapiens", "Homo sapiens"),
    ("Saimiri sciureus", "Primata sp."),
    ("Homo sapiens", "Gnathostomata sp."),
]

INCOMPATIBLE = [
    ("Macaca mulatta", "Macaca fascicularis"),
    ("Carassius gibelio", "Homo sapiens"),
    ("Bos taurus", "Ovis aries"),
    # the ontology is nomenclature-shaped: human hangs off the root, so
    # Primata sp. is not one of its ancestors even though a squirrel monkey's is
    ("Homo sapiens", "Primata sp."),
]


@pytest.mark.parametrize("a,b", COMPATIBLE)
def test_compatible_species(a, b):
    assert Species.get_by_latin_name(a).compatible_with(b)


@pytest.mark.parametrize("a,b", COMPATIBLE)
def test_compatibility_is_symmetric(a, b):
    assert Species.get_by_latin_name(b).compatible_with(a)


@pytest.mark.parametrize("a,b", INCOMPATIBLE)
def test_incompatible_species(a, b):
    assert not Species.get_by_latin_name(a).compatible_with(b)
    assert not Species.get_by_latin_name(b).compatible_with(a)


def test_compatible_with_accepts_prefixes_and_common_names():
    cattle = Species.get("BoLA")
    assert cattle.compatible_with("Bota")
    assert cattle.compatible_with("Bos taurus")
    assert cattle.compatible_with("cow")


def test_compatible_with_returns_false_for_unknown_species():
    assert not Species.get_by_latin_name("Homo sapiens").compatible_with("not a species")


def test_sharing_a_common_ancestor_is_not_the_same_predicate():
    """
    Every species descends from Gnathostomata sp., so a common-ancestor test
    would accept anything. compatible_with requires a direct relation.
    """
    human = Species.get_by_latin_name("Homo sapiens")
    carp = Species.get_by_latin_name("Carassius gibelio")
    root = Species.get_by_latin_name("Gnathostomata sp.")
    assert root.is_ancestor_of(human) and root.is_ancestor_of(carp)
    assert not human.compatible_with(carp)



import pytest

from mhcgnomes import Allele, parse

# One accession-backed representative for every species prefix in the official IPD-MHC 3.8.0.0
# protein release. These explicit cases preserve broad cross-species coverage without deriving
# biological behavior from name regexes. Source commit and checksum are pinned in
# mhcgnomes/data/external_sources.json.
IPD_PREFIX_EXAMPLES = (
    ("Alpi", "IPD-MHC:NHP09575", "Alpi-DRB*W105:01"),
    ("Aoaz", "IPD-MHC:NHP01450", "Aoaz-DRB3*06:01"),
    ("Aole", "IPD-MHC:NHP10004", "Aole-F*05:01"),
    ("Aona", "IPD-MHC:NHP00001", "Aona-DQA1*27:01"),
    ("Aoni", "IPD-MHC:NHP00056", "Aoni-DRB1*03:01"),
    ("Aotr", "IPD-MHC:NHP00058", "Aotr-DQB2*01:02"),
    ("Aovo", "IPD-MHC:NHP01546", "Aovo-DRB1*03:01"),
    ("Atbe", "IPD-MHC:NHP00075", "Atbe-B*01"),
    ("Atfu", "IPD-MHC:NHP00083", "Atfu-E*03:01"),
    ("BoLA", "IPD-MHC:BoLA02982", "BoLA-2*012:01"),
    ("Bofr", "IPD-MHC:BoLA09755", "Bofr-DQB*043:01"),
    ("Bogr", "IPD-MHC:BoLA09194", "Bogr-DQA1*001:01"),
    ("Bubu", "IPD-MHC:BoLA03223", "Bubu-DQA*001:01"),
    ("Cahi", "IPD-MHC:CLA00002", "Cahi-DRB1*02:01:01"),
    ("Caja", "IPD-MHC:NHP00086", "Caja-DQA1*27:01"),
    ("Camo", "IPD-MHC:NHP00122", "Camo-DRB11*01:01"),
    ("Capy", "IPD-MHC:NHP00153", "Capy-DRB1*03:01"),
    ("Ceap", "IPD-MHC:NHP00137", "Ceap-DRB*W013:01"),
    ("Ceat", "IPD-MHC:NHP06535", "Ceat-A1*01:01"),
    ("Cemi", "IPD-MHC:NHP00143", "Cemi-DQA1*01:01"),
    ("Cene", "IPD-MHC:NHP00151", "Cene-DQA1*26:01"),
    ("Chae", "IPD-MHC:NHP00155", "Chae-DQA1*01:01"),
    ("Chpy", "IPD-MHC:NHP09290", "Chpy-A*01:01"),
    ("Chsa", "IPD-MHC:NHP04182", "Chsa-A*04:01"),
    ("Cogu", "IPD-MHC:NHP00178", "Cogu-DQA1*25:01"),
    ("DLA", "IPD-MHC:DLA04814", "DLA-DQA1*012:01:2"),
    ("Eqca", "IPD-MHC:ELA04917", "Eqca-DQA1*001:01"),
    ("Gaga", "IPD-MHC:CHICKEN08554", "Gaga-BF1*002:01:01"),
    ("Gobe", "IPD-MHC:NHP09050", "Gobe-OKO*01:01"),
    ("Gogo", "IPD-MHC:NHP00179", "Gogo-A*01:01:01"),
    ("Hyla", "IPD-MHC:NHP00241", "Hyla-A*01"),
    ("Hymo", "IPD-MHC:NHP05528", "Hymo-DRB*W094:01"),
    ("Lero", "IPD-MHC:NHP00254", "Lero-G*03:01"),
    ("Loat", "IPD-MHC:NHP00256", "Loat-DQA1*01:01"),
    ("Maar", "IPD-MHC:NHP00257", "Maar-DPA1*02:01"),
    ("Maas", "IPD-MHC:NHP05090", "Maas-B*039:01"),
    ("Mafa", "IPD-MHC:NHP00272", "Mafa-DPA1*02:01"),
    ("Mafu", "IPD-MHC:NHP01068", "Mafu-DRB1*10:01"),
    ("Male", "IPD-MHC:NHP04403", "Male-DRB1*03:01"),
    ("Malo", "IPD-MHC:NHP07111", "Malo-A1*003:01"),
    ("Mamu", "IPD-MHC:NHP00361", "Mamu-DPA1*02:01"),
    ("Mane", "IPD-MHC:NHP00610", "Mane-A1*003:01"),
    ("Masi", "IPD-MHC:NHP00634", "Masi-DRB1*03:01"),
    ("Masp", "IPD-MHC:NHP01348", "Masp-DRB1*03:01"),
    ("Math", "IPD-MHC:NHP04436", "Math-DPB1*15:01"),
    ("Onmy", "IPD-MHC:FISH08119", "Onmy-DAA*01:01"),
    ("Ovar", "IPD-MHC:OLA02424", "Ovar-DRB1*03:02"),
    ("Ovca", "IPD-MHC:OLA08844", "Ovca-DRA*01:01:01"),
    ("Paan", "IPD-MHC:NHP00643", "Paan-AG*01:01:01"),
    ("Pacy", "IPD-MHC:NHP00651", "Pacy-A*03:01"),
    ("Paha", "IPD-MHC:NHP00669", "Paha-DPA1*02:01"),
    ("Papa", "IPD-MHC:NHP00678", "Papa-A*01:01"),
    ("Papp", "IPD-MHC:NHP00701", "Papp-DQA1*01:01"),
    ("Patr", "IPD-MHC:NHP00705", "Patr-A*01:01"),
    ("Paur", "IPD-MHC:NHP01381", "Paur-DRB*W048:01"),
    ("Pipi", "IPD-MHC:NHP00913", "Pipi-B*01"),
    ("Poab", "IPD-MHC:NHP04649", "Poab-B*04:01:01:01"),
    ("Popy", "IPD-MHC:NHP00922", "Popy-A*01:01"),
    ("Rano", "IPD-MHC:RT108344", "Rano-A*av1"),
    ("SLA", "IPD-MHC:SLA05920", "SLA-DMA*01:01:01"),
    ("Safu", "IPD-MHC:NHP00966", "Safu-G*03:01"),
    ("Sage", "IPD-MHC:NHP00972", "Sage-PS1*01"),
    ("Sala", "IPD-MHC:NHP04229", "Sala-DRB*W076:01"),
    ("Samy", "IPD-MHC:NHP00974", "Samy-PS1*01"),
    ("Saoe", "IPD-MHC:NHP00976", "Saoe-DPB1*01:01"),
    ("Sasa", "IPD-MHC:FISH08192", "Sasa-DAA*01:01"),
    ("Sasc", "IPD-MHC:NHP01052", "Sasc-DPA1*05:01"),
    ("Seen", "IPD-MHC:NHP02068", "Seen-DPB1*01"),
    ("Thge", "IPD-MHC:NHP02069", "Thge-DQA1*01:01"),
)


def test_ipd_prefix_inventory_is_explicit_and_cross_species():
    prefixes = [prefix for prefix, _accession, _raw_string in IPD_PREFIX_EXAMPLES]
    assert len(prefixes) == 69
    assert len(set(prefixes)) == len(prefixes)
    assert {"DLA", "Gaga", "Onmy", "SLA", "Mamu"}.issubset(prefixes)


@pytest.mark.parametrize("prefix, accession, raw_string", IPD_PREFIX_EXAMPLES)
def test_ipd_mhc_short_prefix_example_parses_without_species(prefix, accession, raw_string):
    result = parse(raw_string, raise_on_error=False)
    assert isinstance(result, Allele), (
        f"Failed to parse IPD-MHC {accession} ({prefix}) as an allele: {raw_string}; got {result!r}"
    )



from pathlib import Path

import pytest
from release_utils import (
    ReleaseError,
    check_tag_against_version_file,
    ensure_tag_matches_version,
    main,
    parse_version_from_python_file,
    tag_for_version,
)


def write_version_file(tmp_path: Path, contents: str) -> Path:
    path = tmp_path / "version.py"
    path.write_text(contents, encoding="utf-8")
    return path


def test_parse_version_from_python_file_assign(tmp_path: Path):
    path = write_version_file(tmp_path, '__version__ = "3.15.1"\n')
    assert parse_version_from_python_file(path) == "3.15.1"


def test_parse_version_from_python_file_annotated_assign(tmp_path: Path):
    path = write_version_file(tmp_path, '__version__: str = "3.15.1rc1"\n')
    assert parse_version_from_python_file(path) == "3.15.1rc1"


def test_parse_version_from_python_file_rejects_multiple_versions(tmp_path: Path):
    path = write_version_file(tmp_path, '__version__ = "1.0.0"\n__version__ = "1.0.1"\n')
    with pytest.raises(ReleaseError, match="Multiple __version__ assignments"):
        parse_version_from_python_file(path)


def test_parse_version_from_python_file_rejects_non_literal(tmp_path: Path):
    path = write_version_file(tmp_path, "__version__ = VERSION\n")
    with pytest.raises(ReleaseError, match="string literal"):
        parse_version_from_python_file(path)


def test_tag_for_version():
    assert tag_for_version("3.15.1") == "v3.15.1"


def test_ensure_tag_matches_version_accepts_matching_tag():
    ensure_tag_matches_version("v3.15.1", "3.15.1")


def test_ensure_tag_matches_version_rejects_mismatch():
    with pytest.raises(ReleaseError, match="Tag/version mismatch"):
        ensure_tag_matches_version("v3.15.2", "3.15.1")


def test_check_tag_against_version_file(tmp_path: Path):
    path = write_version_file(tmp_path, '__version__ = "3.15.1"\n')
    assert check_tag_against_version_file("v3.15.1", path) == "3.15.1"


def test_release_utils_main_print_version(tmp_path: Path, capsys):
    path = write_version_file(tmp_path, '__version__ = "3.15.1"\n')
    assert main(["print-version", "--version-file", str(path), "--as-tag"]) == 0
    captured = capsys.readouterr()
    assert captured.out.strip() == "v3.15.1"


def test_release_utils_main_check_tag_failure(tmp_path: Path, capsys):
    path = write_version_file(tmp_path, '__version__ = "3.15.1"\n')
    assert main(["check-tag", "--version-file", str(path), "--tag", "v3.15.2"]) == 1
    captured = capsys.readouterr()
    assert "Tag/version mismatch" in captured.err



#!/usr/bin/env python3
"""
Systematic PubMed search for MHC papers with downloadable allele lists.

Runs one query per taxonomic stratum, retrieves paper metadata,
and outputs a TSV of candidate papers for manual review.

Usage:
    python paper/scripts/collect_pubmed.py --output paper/validation/candidate_papers.tsv
"""

import argparse
import csv
import sys
import time

try:
    from Bio import Entrez
except ImportError:
    print("Requires biopython: pip install biopython", file=sys.stderr)
    sys.exit(1)

Entrez.email = "mhcgnomes-paper@example.com"

# One query per taxonomic stratum, targeting papers with supplementary
# allele data published 2015-2025
STRATA = {
    "human_clinical": (
        '("HLA typing" OR "HLA genotyping") '
        "AND (supplementary OR table) "
        "AND 2015:2025[dp] "
        "NOT review[pt]"
    ),
    "human_population": (
        '("HLA frequency" OR "HLA distribution" OR "allele frequency") '
        "AND population AND HLA "
        "AND 2015:2025[dp] "
        "NOT review[pt]"
    ),
    "nonhuman_primate": (
        "(MHC OR Mamu OR Mafa OR Patr) "
        "AND (macaque OR chimpanzee OR marmoset) "
        "AND (genotyping OR allele) "
        "AND 2015:2025[dp]"
    ),
    "rodent": (
        '(MHC OR "H-2" OR H2) '
        "AND (mouse OR rat OR hamster) "
        "AND (haplotype OR genotyping OR allele) "
        "AND 2015:2025[dp] "
        "NOT review[pt]"
    ),
    "bird": (
        '("MHC diversity" OR "MHC class II" OR "MHC genotyping") '
        "AND (passerine OR raptor OR penguin OR waterfowl OR songbird OR avian) "
        "AND 2018:2025[dp]"
    ),
    "fish": (
        "(MHC) AND (salmonid OR cichlid OR zebrafish OR carp OR teleost) "
        "AND (diversity OR genotyping OR class) "
        "AND 2018:2025[dp]"
    ),
    "reptile_amphibian": (
        "(MHC) AND (turtle OR lizard OR snake OR frog OR amphibian OR reptile) "
        'AND ("class I" OR "class II" OR diversity) '
        "AND 2018:2025[dp]"
    ),
    "livestock": (
        '(MHC OR BoLA OR SLA OR OLA OR ELA) AND (genotyping OR typing OR allele) '
        "AND (cattle OR pig OR sheep OR horse) "
        "AND 2018:2025[dp] "
        "NOT review[pt]"
    ),
    "wildlife_conservation": (
        '("MHC diversity") '
        "AND (endangered OR conservation OR bottleneck OR threatened) "
        "AND 2018:2025[dp]"
    ),
}


def search_pubmed(query, max_results=20):
    """Search PubMed and return PMIDs."""
    try:
        handle = Entrez.esearch(
            db="pubmed", term=query, retmax=max_results, sort="relevance"
        )
        result = Entrez.read(handle)
        handle.close()
        return result["IdList"]
    except Exception as e:
        print(f"  Search failed: {e}", file=sys.stderr)
        return []


def fetch_paper_metadata(pmids):
    """Fetch title, authors, journal, year, DOI for a list of PMIDs."""
    if not pmids:
        return []
    try:
        handle = Entrez.efetch(
            db="pubmed", id=pmids, rettype="xml", retmode="xml"
        )
        records = Entrez.read(handle)
        handle.close()
    except Exception as e:
        print(f"  Fetch failed: {e}", file=sys.stderr)
        return []

    papers = []
    for article in records.get("PubmedArticle", []):
        medline = article.get("MedlineCitation", {})
        art = medline.get("Article", {})

        pmid = str(medline.get("PMID", ""))
        title = art.get("ArticleTitle", "")

        # Extract year
        journal = art.get("Journal", {})
        ji = journal.get("JournalIssue", {})
        pd = ji.get("PubDate", {})
        year = pd.get("Year", pd.get("MedlineDate", ""))

        journal_name = journal.get("Title", "")

        # Extract DOI
        doi = ""
        for eid in art.get("EIdList", []):
            if hasattr(eid, "attributes") and eid.attributes.get("EIdType") == "doi":
                doi = str(eid)
        if not doi:
            article_ids = article.get("PubmedData", {}).get("ArticleIdList", [])
            for aid in article_ids:
                if hasattr(aid, "attributes") and aid.attributes.get("IdType") == "doi":
                    doi = str(aid)

        # First author
        author_list = art.get("AuthorList", [])
        first_author = ""
        if author_list:
            a = author_list[0]
            last = a.get("LastName", "")
            init = a.get("Initials", "")
            first_author = f"{last} {init}" if last else ""

        papers.append({
            "pmid": pmid,
            "first_author": first_author,
            "year": year,
            "title": title,
            "journal": journal_name,
            "doi": doi,
        })

    return papers


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--output", default="paper/validation/candidate_papers.tsv"
    )
    parser.add_argument(
        "--max-per-stratum", type=int, default=10,
        help="Max papers to retrieve per stratum"
    )
    args = parser.parse_args()

    all_papers = []

    for stratum, query in STRATA.items():
        print(f"Searching stratum: {stratum}", file=sys.stderr)
        pmids = search_pubmed(query, max_results=args.max_per_stratum)
        print(f"  Found {len(pmids)} papers", file=sys.stderr)

        if pmids:
            papers = fetch_paper_metadata(pmids)
            for p in papers:
                p["stratum"] = stratum
            all_papers.extend(papers)
            time.sleep(0.5)

    print(f"\nTotal: {len(all_papers)} candidate papers", file=sys.stderr)

    with open(args.output, "w") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=[
                "stratum", "pmid", "first_author", "year",
                "title", "journal", "doi",
            ],
            delimiter="\t",
        )
        writer.writeheader()
        writer.writerows(all_papers)

    print(f"Written to {args.output}", file=sys.stderr)


if __name__ == "__main__":
    main()



import hashlib
import json
import subprocess
import sys

from mhcgnomes.version import __version__


def run_cli(*args):
    return subprocess.run(
        [sys.executable, "-m", "mhcgnomes", *args],
        capture_output=True,
        text=True,
        check=False,
    )


def run_cli_with_stdin(stdin_text, *args):
    return subprocess.run(
        [sys.executable, "-m", "mhcgnomes", *args],
        input=stdin_text,
        capture_output=True,
        text=True,
        check=False,
    )


def test_cli_help():
    result = run_cli("--help")
    assert result.returncode == 0
    assert "usage:" in result.stdout.lower()
    assert "mhcgnomes" in result.stdout.lower()
    assert "mhcgnomes data" in result.stdout


def test_cli_version():
    result = run_cli("--version")

    assert result.returncode == 0
    assert result.stdout.strip() == f"mhcgnomes {__version__}"
    assert result.stderr == ""


def test_data_command_lists_sources_by_default():
    result = run_cli("data")

    assert result.returncode == 0
    assert "imgt-hla-allele-history-3.42.0" in result.stdout
    assert "ipd-mhc-3.8.0.0-protein" in result.stdout


def test_data_list_is_explicit_alias_for_default_action():
    default = run_cli("data")
    explicit = run_cli("data", "list")

    assert explicit.returncode == 0
    assert explicit.stdout == default.stdout


def test_data_download_help_owns_download_options():
    result = run_cli("data", "download", "--help")

    assert result.returncode == 0
    assert "--cache-dir" in result.stdout
    assert "--destination" in result.stdout
    assert "--mirror" in result.stdout
    assert "--offline" in result.stdout


def test_data_download_options_are_not_accepted_by_data_group():
    result = run_cli("data", "--offline")

    assert result.returncode == 2
    assert "unrecognized arguments: --offline" in result.stderr


def test_data_download_materializes_from_local_mirror_offline(tmp_path):
    payload = b"nested CLI payload"
    digest = hashlib.sha256(payload).hexdigest()
    manifest = tmp_path / "manifest.json"
    manifest.write_text(
        json.dumps(
            {
                "schema_version": 1,
                "sources": [
                    {
                        "id": "nested-cli-source",
                        "filename": "nested.dat",
                        "provider": "Example",
                        "release": "1",
                        "url": "https://official.example/nested.dat",
                        "sha256": digest,
                        "license_url": "https://official.example/license",
                        "groups": ["example"],
                        "default": True,
                    }
                ],
            }
        ),
        encoding="utf-8",
    )
    mirror = tmp_path / "mirror"
    mirror.mkdir()
    (mirror / "nested.dat").write_bytes(payload)
    destination = tmp_path / "destination"

    result = run_cli(
        "data",
        "download",
        "--manifest",
        str(manifest),
        "--cache-dir",
        str(tmp_path / "cache"),
        "--destination",
        str(destination),
        "--mirror",
        str(mirror),
        "--offline",
    )

    assert result.returncode == 0
    assert (destination / "nested.dat").read_bytes() == payload
    assert "nested-cli-source" in result.stdout


def test_cli_table_output_for_valid_allele():
    result = run_cli("HLA-A*02:01")
    assert result.returncode == 0
    output = result.stdout
    assert "input" in output
    assert "type" in output
    assert "normalized" in output
    assert "properties" in output
    assert "HLA-A*02:01" in output
    assert "Allele" in output


def test_cli_tsv_output_multiple_inputs():
    result = run_cli("--format", "tsv", "HLA-A*02:01", "DQ2.5")
    assert result.returncode == 0
    lines = [line for line in result.stdout.strip().splitlines() if line]
    assert len(lines) == 3
    assert lines[0].split("\t") == [
        "input",
        "type",
        "normalized",
        "compact",
        "species",
        "gene",
        "mhc_class",
        "properties",
    ]
    assert "HLA-A*02:01" in lines[1]
    assert "DQ2.5" in lines[2]


def test_cli_strict_mode_fails_on_unparseable_input():
    result = run_cli("--strict", "NOT_A_REAL_ALLELE")
    assert result.returncode == 1
    assert "ParseError" in result.stderr


def test_cli_non_strict_shows_parse_error_row():
    result = run_cli("--format", "tsv", "NOT_A_REAL_ALLELE")
    assert result.returncode == 0
    assert "ParseError" in result.stdout


def test_cli_reads_names_from_stdin_when_no_positional_args_are_given():
    result = run_cli_with_stdin("HLA-A*02:01\nDQ2.5\n", "--format", "json")

    assert result.returncode == 0
    rows = json.loads(result.stdout)
    assert [row["input"] for row in rows] == ["HLA-A*02:01", "DQ2.5"]
    assert rows[0]["normalized"] == "HLA-A*02:01"
    assert rows[1]["type"] == "Pair"


def test_cli_no_header_omits_tsv_header_row():
    result = run_cli("--format", "tsv", "--no-header", "HLA-A*02:01")

    assert result.returncode == 0
    lines = [line for line in result.stdout.strip().splitlines() if line]
    assert len(lines) == 1
    assert not lines[0].startswith("input\t")
    assert lines[0].startswith("HLA-A*02:01\tAllele\t")


def test_cli_errors_when_no_input_is_provided():
    result = run_cli_with_stdin("")

    assert result.returncode == 2
    assert "Provide one or more names" in result.stderr



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Union

from .result_with_mhc_class import ResultWithMhcClass
from .species import Species


@dataclass(eq=False, repr=False, frozen=True, init=False)
class GeneClassInfo(ResultWithMhcClass):
    """
    Lenient species-aware gene classification result.

    Unlike ``Gene`` or ``Allele``, this object can represent heuristic
    class/chain inference for source-backed gene tokens that are not yet part of
    the strict runtime ontology.
    """

    gene_name: Union[str, None] = None
    chain: Union[str, None] = None
    non_mhc: bool = False
    source: str = ""

    def __init__(
        self,
        species: Species,
        mhc_class: Union[str, None],
        gene_name: Union[str, None] = None,
        chain: Union[str, None] = None,
        non_mhc: bool = False,
        source: str = "",
        raw_string: Union[str, None] = None,
    ):
        ResultWithMhcClass.__init__(
            self, species=species, mhc_class=mhc_class, raw_string=raw_string
        )
        self._set_field(self, "gene_name", gene_name)
        self._set_field(self, "chain", chain)
        self._set_field(self, "non_mhc", bool(non_mhc))
        self._set_field(self, "source", source)

    def to_string(self, include_species=True, use_old_species_prefix=False):
        prefix = ""
        if include_species:
            if use_old_species_prefix:
                prefix = f"{self.species.historic_alias}-"
            else:
                prefix = f"{self.species.prefix}-"

        if self.gene_name:
            return f"{prefix}{self.gene_name}"

        class_bits = []
        if self.mhc_class:
            class_bits.append(f"class {self.mhc_class}")
        if self.chain:
            class_bits.append(self.chain)
        if self.non_mhc and not class_bits:
            class_bits.append("non-MHC")
        label = " ".join(class_bits) if class_bits else "unknown"
        if include_species:
            return f"{self.species.prefix} {label}"
        return label

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    def to_record(self):
        record = self.species.to_record()
        if self.gene_name is not None:
            record["gene_name"] = self.gene_name
        if self.mhc_class is not None:
            record["mhc_class"] = self.mhc_class
        if self.chain is not None:
            record["chain"] = self.chain
        record["non_mhc"] = self.non_mhc
        if self.source:
            record["source"] = self.source
        return record



from pathlib import Path

import pytest

from mhcgnomes.allele import Allele
from mhcgnomes.parser import Parser
from mhcgnomes.serotype import Serotype


@pytest.mark.parametrize(
    ("name", "expected"),
    [
        ("HLA-A02:01", "HLA-A*02:01"),
        ("Patr-A0101", "Patr-A*01:01"),
        ("Mamu-A1:00101", "Mamu-A1*001:01"),
        ("SLA-1:0101", "SLA-1*01:01"),
        ("BoLA-1:00901", "BoLA-1*009:01"),
        ("DLA-8803401", "DLA-88*034:01"),
        ("Eqca-16:00101", "Eqca-16*001:01"),
        ("Gogo-B0101", "Gogo-B*01:01"),
    ],
)
def test_explicit_species_allele_fast_path_is_cross_species(monkeypatch, name, expected):
    parser = Parser()

    def fail_if_general_parser_runs(*args, **kwargs):
        raise AssertionError("eligible allele unexpectedly used the general parser")

    monkeypatch.setattr(parser, "parse_multiple_candidates", fail_if_general_parser_runs)

    result = parser.parse(name)

    assert result.to_string() == expected
    assert result.raw_string == name


@pytest.mark.parametrize(
    "name",
    [
        "A*02:01",
        "HLA-A2",
        "BoLA-HD6",
        "H-2-Db",
        "Gaga-B11",
        "Gaga-BF19",
        "BoLA-DRA-DRB31501",
        "HLA-DRA*01:01/DRB1*03:01",
        "Caja-PS*02:01",
    ],
)
def test_fast_path_defers_unprefixed_non_allele_and_complex_names(name):
    assert Parser()._parse_explicit_species_allele_candidates(name) == []


def test_fast_path_preserves_ambiguous_candidates_for_result_type_filters():
    parser = Parser()

    assert type(parser.parse("HLA-B3901")) is Allele
    assert type(parser.parse("HLA-B3901", preferred_result_types=Serotype)) is Serotype
    assert type(parser.parse("HLA-B3901", required_result_types=Serotype)) is Serotype


@pytest.mark.parametrize(
    "parse_kwargs",
    [
        {"preferred_result_types": Allele},
        {"required_result_types": Allele},
        {"only_class1": True},
        {"only_class2": True, "raise_on_error": False},
    ],
)
def test_fast_path_defers_to_general_parser_for_candidate_filters(monkeypatch, parse_kwargs):
    parser = Parser()
    general_parser_calls = []
    parse_multiple_candidates = parser.parse_multiple_candidates

    def track_general_parser(*args, **kwargs):
        general_parser_calls.append((args, kwargs))
        return parse_multiple_candidates(*args, **kwargs)

    monkeypatch.setattr(parser, "parse_multiple_candidates", track_general_parser)

    parser.parse("HLA-B3901", **parse_kwargs)

    assert len(general_parser_calls) == 1


def test_fast_path_matches_general_parser_for_checked_in_netmhcpan_inventory():
    inventory_path = Path(__file__).parents[1] / "evaluation" / "netmhcpan-4.1-alleles.txt"
    with inventory_path.open() as inventory_file:
        names = [line.split("\t", 1)[0].strip() for line in inventory_file if line.strip()]

    fast_parser = Parser()
    general_parser = Parser()
    general_parser._parse_explicit_species_allele_candidates = lambda name: []

    fast_results = [fast_parser.parse(name) for name in names]
    general_results = [general_parser.parse(name) for name in names]

    assert len(names) == 11_114
    assert fast_results == general_results
    assert [result.to_string() for result in fast_results] == [
        result.to_string() for result in general_results
    ]



import pytest

from mhcgnomes import (
    Allele,
    Class2Locus,
    Gene,
    Haplotype,
    Pair,
    ParseError,
    parse,
)

from .common import eq_


def test_mouse_class1_alleles_H2_Kk():
    H2Kk = Allele.get("H2", "K", "k")

    eq_(parse("H2-Kk"), H2Kk)

    # with a hyphen in "H-2"
    eq_(parse("H-2-Kk"), H2Kk)


def test_mouse_class1_alleles_H2_Db():
    H2Db = Allele.get("H2", "D", "b")

    eq_(parse("H2-Db"), H2Db)

    # with hyphen in "H-2"
    eq_(parse("H-2-Db"), H2Db)


def test_H2_Kd_without_seps():
    eq_(parse("H2Kd"), Allele.get("H2", "K", "d"))


def test_H2_Lq_with_dash_in_species():
    eq_(parse("H-2-Lq"), Allele.get("H2", "L", "q"))


def test_H2_Lq_without_dash_in_species():
    eq_(parse("H2-Lq"), Allele.get("H2", "L", "q"))


def test_mouse_class2_gene():
    # H2-IAb
    gene = Gene.get("H2", "EB2")
    eq_(parse("H2-IEb2"), gene)

    # with hyphen in "H-2"
    eq_(parse("H-2-IEb2"), gene)


def test_parse_H2r():
    haplotype = parse("H2-r")
    assert isinstance(haplotype, Haplotype)
    eq_(haplotype.to_string(), "H2-r")


def test_parse_H2_IE():
    result = parse("H2-IE")
    eq_(type(result), Class2Locus)
    eq_(result.name, "E")


def test_mouse_MR1_weird_uniprot_entry():
    seq = "Major histocompatibility complex class I-related gene protein OS=Mus musculus OX=10090 GN=Mr1 PE=1 SV=2"
    result = parse(seq)
    expected = Gene.get("H2", "MR1")
    eq_(result, expected)


def test_parse_H2_IEd_simplify():
    result = parse("H2-IEd", collapse_singleton_haplotypes=True)
    eq_(type(result), Pair)
    eq_(result.alpha.name, "d")
    eq_(result.beta.name, "d")


def test_parse_H2_IEd_no_simplify():
    result = parse("H2-IEd", collapse_singleton_haplotypes=False)
    eq_(type(result), Haplotype)
    eq_(result.name, "d")
    assert result.locus_restriction is not None


# ---- Mouse class II gene name variants ----


def test_parse_mouse_MHC_II_IE_beta():
    """'mouse MHC II IE-beta' should parse as H2-EB gene."""
    result = parse("mouse MHC II IE-beta")
    assert result is not None, "Failed to parse 'mouse MHC II IE-beta'"
    assert isinstance(result, Gene)
    eq_(result.gene_name, "EB")
    eq_(result.species_prefix, "H2")


def test_parse_MHC_II_I_E_beta():
    """'MHC II I-E beta' should parse as H2-EB gene."""
    result = parse("MHC II I-E beta")
    assert result is not None, "Failed to parse 'MHC II I-E beta'"
    assert isinstance(result, Gene)
    eq_(result.gene_name, "EB")


def test_parse_murine_MHC_class_II_I_E_beta():
    """'murine MHC class II I-E beta' should parse as H2-EB gene."""
    result = parse("murine MHC class II I-E beta")
    assert result is not None, "Failed to parse 'murine MHC class II I-E beta'"
    assert isinstance(result, Gene)
    eq_(result.gene_name, "EB")
    eq_(result.species_prefix, "H2")


def test_parse_H2_IE_beta_chain():
    """'H2-IE beta chain' should parse as H2-EB gene."""
    result = parse("H2-IE beta chain")
    assert result is not None, "Failed to parse 'H2-IE beta chain'"
    assert isinstance(result, Gene)
    eq_(result.gene_name, "EB")


def test_parse_I_E_beta_chain():
    """'I-E beta chain' should parse as H2-EB gene."""
    result = parse("I-E beta chain")
    assert result is not None, "Failed to parse 'I-E beta chain'"
    assert isinstance(result, Gene)
    eq_(result.gene_name, "EB")


def test_parse_murine_species():
    """'murine' should be recognized as Mus musculus."""
    from mhcgnomes import Species

    species = Species.get("murine")
    assert species is not None
    eq_(species.prefix, "H2")


# ---- Contradictory class markers should fail ----


def test_mouse_MHC_I_IE_beta_fails():
    """IE is class II, so 'mouse MHC I IE-beta' is contradictory and must not parse."""
    with pytest.raises(ParseError):
        parse("mouse MHC I IE-beta")


def test_mouse_MHC_class_I_I_A_alpha_fails():
    """I-A is class II, so 'mouse MHC class I I-A alpha' is contradictory."""
    with pytest.raises(ParseError):
        parse("mouse MHC class I I-A alpha")


def test_mouse_MHC_class_I_I_E_beta_fails():
    """I-E is class II, so 'mouse MHC class I I-E beta' is contradictory."""
    with pytest.raises(ParseError):
        parse("mouse MHC class I I-E beta")



from mhcgnomes.validate_external_names import fasta_names, main, validate_fasta


def test_fasta_names_reads_provider_header_name_field():
    lines = [
        ">HLA:HLA00001 A*01:01:01:01 365 bp\n",
        "AAAA\n",
        ">IPD-MHC:DLA04814 DLA-DQA1*012:01:2 77 bp\n",
    ]
    assert list(fasta_names(lines)) == [(1, "A*01:01:01:01"), (3, "DLA-DQA1*012:01:2")]


def test_validate_fasta_checks_human_and_cross_species_names(tmp_path):
    path = tmp_path / "protein.fasta"
    path.write_text(
        ">HLA:HLA00001 A*01:01:01:01 365 bp\n"
        "AAAA\n"
        ">IPD-MHC:DLA04814 DLA-DQA1*012:01:2 77 bp\n"
        "AAAA\n"
        ">IPD-MHC:FISH08119 Onmy-DAA*01:01 74 bp\n"
        "AAAA\n"
        ">IPD-MHC:SLA05920 SLA-DMA*01:01:01 92 bp\n"
        "AAAA\n",
        encoding="utf-8",
    )

    result = validate_fasta(path)

    assert result.record_count == 4
    assert result.unique_name_count == 4
    assert result.failures == ()


def test_validate_fasta_reports_malformed_and_unparseable_headers(tmp_path):
    path = tmp_path / "bad.fasta"
    path.write_text(">missing-name\nAAAA\n>DB not-an-allele\nAAAA\n", encoding="utf-8")

    result = validate_fasta(path)

    assert result.record_count == 2
    assert result.unique_name_count == 1
    assert len(result.failures) == 2
    assert "malformed FASTA header" in result.failures[0]
    assert "not-an-allele" in result.failures[1]


def test_validate_fasta_rejects_file_without_headers(tmp_path):
    path = tmp_path / "empty.fasta"
    path.write_text("AAAA\n", encoding="utf-8")

    result = validate_fasta(path)

    assert result.failures == ("no FASTA headers found",)


def test_cli_reports_missing_source(tmp_path, capsys):
    missing = tmp_path / "missing.fasta"

    assert main([str(missing)]) == 1
    assert "mhcgnomes data download --group validation" in capsys.readouterr().err



import json
import os
import subprocess
import sys


def test_cli_parses_bundled_data_under_ascii_locale():
    env = os.environ.copy()
    env.update(
        {
            "LANG": "C",
            "LC_ALL": "C",
            "PYTHONCOERCECLOCALE": "0",
            "PYTHONUTF8": "0",
        }
    )

    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "mhcgnomes",
            "--format",
            "json",
            "HLA-A*02:01",
        ],
        capture_output=True,
        text=True,
        check=False,
        env=env,
    )

    assert result.returncode == 0, result.stderr
    rows = json.loads(result.stdout)
    assert len(rows) == 1
    assert rows[0]["input"] == "HLA-A*02:01"
    assert rows[0]["type"] == "Allele"
    assert rows[0]["normalized"] == "HLA-A*02:01"
    assert rows[0]["compact"] == "A0201"
    assert rows[0]["species"] == "HLA"
    assert rows[0]["gene"] == "A"
    assert rows[0]["mhc_class"] == "Ia"
    assert "species_name=Homo sapiens" in rows[0]["properties"]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
from pathlib import Path


def test_test_sh_uses_python_module_for_pytest(tmp_path):
    marker = tmp_path / "python-pytest-called"
    fake_bin = tmp_path / "bin"
    fake_bin.mkdir()

    fake_python = fake_bin / "python"
    fake_python.write_text(
        "#!/bin/sh\n"
        'if [ "$1" = "-c" ]; then exit 1; fi\n'
        'if [ "$1" = "-m" ] && [ "$2" = "pytest" ]; then\n'
        '  touch "$MHCGENOMES_TEST_MARKER"\n'
        "  exit 0\n"
        "fi\n"
        "exit 2\n"
    )
    fake_python.chmod(0o755)

    fake_pytest = fake_bin / "pytest"
    fake_pytest.write_text("#!/bin/sh\nexit 97\n")
    fake_pytest.chmod(0o755)

    repo_dir = Path(__file__).resolve().parents[1]
    env = os.environ.copy()
    env["PATH"] = os.pathsep.join((str(fake_bin), env["PATH"]))
    env["MHCGENOMES_TEST_MARKER"] = str(marker)
    result = subprocess.run(
        ["bash", "test.sh", "--collect-only"],
        cwd=repo_dir,
        env=env,
        capture_output=True,
        text=True,
        check=False,
    )

    assert result.returncode == 0, result.stderr
    assert marker.exists()



import pytest

from mhcgnomes import (
    MhcClass,
    parse,
)


@pytest.mark.parametrize(
    "s",
    [
        "human class 1",
        "human class i",
        "human class I",
        "hla class 1",
        "hla class i",
        "hla class I",
        "HLA class 1",
        "HLA class i",
        "HLA class I",
        "MHC I",
        "MHC 1",
    ],
)
def test_human_class_1(s):
    """Test parsing of human class I MHC strings."""
    expected_parsed_result = MhcClass.get("HLA", "I")
    expected_string_repr = "human class I"
    parsed_result = parse(s)
    assert parsed_result == expected_parsed_result, (
        f"Expected {expected_parsed_result} for parsing of '{s}' but got {parsed_result}"
    )
    normalized_str = parsed_result.to_string()
    assert normalized_str == expected_string_repr, (
        f"Expected '{expected_string_repr}' for normalized representation of '{s}' "
        f"but got '{normalized_str}'"
    )
    compact_str = parsed_result.compact_string()
    assert compact_str == expected_string_repr, (
        f"Expected '{expected_string_repr}' for compact representation of '{s}' "
        f"but got '{compact_str}'"
    )


@pytest.mark.parametrize(
    "s",
    [
        "MHC II",
        "MHC 2",
        "MHC class II",
        "MHC-II",
    ],
)
def test_mhc_class_2(s):
    """Test that 'MHC II' and variants parse as MhcClass, not Haplotype."""
    result = parse(s)
    assert isinstance(result, MhcClass), (
        f"Expected MhcClass for '{s}' but got {type(result).__name__}: {result}"
    )
    assert result.is_class2


def test_mhc_class_get_rejects_unknown_species_and_invalid_class():
    assert MhcClass.get("NOT_A_SPECIES", "I") is None
    assert MhcClass.get("HLA", "not a class") is None


def test_mhc_class_genes_returns_matching_gene_objects():
    result = MhcClass.get("HLA", "I")

    genes = result.genes()
    gene_names = {gene.name for gene in genes}

    assert "A" in gene_names
    assert "B" in gene_names
    assert "DRA" not in gene_names
    assert all(gene.species_prefix == "HLA" for gene in genes)
    assert all(gene.is_class1 for gene in genes)


def test_mhc_class_to_record_and_strings():
    result = MhcClass.get("HLA", "II")

    assert result is not None
    assert result.is_class2
    assert not result.is_class1
    assert result.to_string() == "human class II"
    assert result.to_string(include_species=False) == "class II"
    assert result.compact_string(include_species=False) == "class II"
    assert result.to_record() == {
        "species_prefix": "HLA",
        "species_name": "Homo sapiens",
        "species_latin_name": "Homo sapiens",
        "mhc_class": "II",
    }



import re
from pathlib import Path

import yaml

from mhcgnomes import Species

SPECIES_PATH = Path(__file__).parent.parent / "mhcgnomes" / "data" / "species.yaml"
GENE_ALIASES_PATH = Path(__file__).parent.parent / "mhcgnomes" / "data" / "gene_aliases.yaml"

# These are unusual but source-backed canonical names already exercised by tests
# and external nomenclature datasets.
ALLOWLISTED_MULTI_LOWERCASE_GENES = {
    "B-F-minor",
    "DQA2like",
    "DQB2like",
    "Mill1",
    "Mill2",
}

ALLOWLISTED_LIKE_GENES = {
    "DQA2like",
    "DQB2like",
}

FORBIDDEN_GENE_PATTERNS = (
    re.compile(r"^LOC\d+$", re.IGNORECASE),
    re.compile(r"^ENS[A-Z0-9]*\d+$", re.IGNORECASE),
)

FORBIDDEN_GENE_SUBSTRINGS = (
    "scaffold",
    "unplaced",
    "protein",
    "chain",
)


def load_yaml(path: Path):
    with path.open() as fd:
        return yaml.safe_load(fd)


def iter_raw_canonical_genes():
    species_data = load_yaml(SPECIES_PATH)
    for latin_name, species_info in species_data.items():
        genes = species_info.get("genes", {}) or {}
        for mhc_class, members in genes.items():
            if isinstance(members, list):
                for gene in members:
                    yield latin_name, mhc_class, None, str(gene)
            else:
                for locus, locus_genes in members.items():
                    for gene in locus_genes:
                        yield latin_name, mhc_class, str(locus), str(gene)


def test_canonical_genes_do_not_look_like_model_records_or_free_text():
    bad = []
    for latin_name, mhc_class, locus, gene in iter_raw_canonical_genes():
        lower = gene.lower()
        if any(ch.isspace() for ch in gene):
            bad.append((latin_name, mhc_class, locus, gene, "whitespace"))
            continue
        if lower.startswith("mhc"):
            bad.append((latin_name, mhc_class, locus, gene, "mhc-prefix"))
            continue
        if any(pattern.match(gene) for pattern in FORBIDDEN_GENE_PATTERNS):
            bad.append((latin_name, mhc_class, locus, gene, "model-record"))
            continue
        for substring in FORBIDDEN_GENE_SUBSTRINGS:
            if substring in lower:
                bad.append((latin_name, mhc_class, locus, gene, substring))
                break
    assert not bad, bad


def test_only_allowlisted_canonical_genes_have_long_lowercase_runs():
    nonstandard = sorted(
        {
            gene
            for _latin_name, _mhc_class, _locus, gene in iter_raw_canonical_genes()
            if re.search(r"[a-z]{3,}", gene)
        }
    )
    assert nonstandard == sorted(ALLOWLISTED_MULTI_LOWERCASE_GENES)


def test_only_allowlisted_canonical_genes_use_like_suffix_style_names():
    like_genes = sorted(
        {
            gene
            for _latin_name, _mhc_class, _locus, gene in iter_raw_canonical_genes()
            if "like" in gene.lower()
        }
    )
    assert like_genes == sorted(ALLOWLISTED_LIKE_GENES)


def test_gene_alias_targets_resolve_to_runtime_gene_or_locus():
    alias_data = load_yaml(GENE_ALIASES_PATH)
    bad = []
    for species_key, aliases in alias_data.items():
        species = Species.get(species_key)
        if species is None:
            bad.append((species_key, None, None, "unknown species key"))
            continue
        for alias, target in aliases.items():
            gene_target = species.find_matching_gene_name(target)
            locus_target = species.find_matching_class2_locus_name(target)
            if gene_target is None and locus_target is None:
                bad.append((species_key, alias, target, "unresolved target"))
    assert not bad, bad


def test_gene_alias_entries_do_not_use_model_record_style_strings():
    alias_data = load_yaml(GENE_ALIASES_PATH)
    bad = []
    for species_key, aliases in alias_data.items():
        for alias, target in aliases.items():
            for role, value in (("alias", alias), ("target", target)):
                lower = value.lower()
                if any(ch.isspace() for ch in value):
                    bad.append((species_key, role, value, "whitespace"))
                    continue
                if any(pattern.match(value) for pattern in FORBIDDEN_GENE_PATTERNS):
                    bad.append((species_key, role, value, "model-record"))
                    continue
                for substring in FORBIDDEN_GENE_SUBSTRINGS:
                    if substring in lower:
                        bad.append((species_key, role, value, substring))
                        break
    assert not bad, bad



import pytest

from mhcgnomes import Haplotype, ParseError, parse

from .common import eq_, raises


@raises(ParseError)
def test_parse_extra_text_after_allele():
    parse("HLA-A*02:01 zipper")


@raises(ParseError)
def test_bad_input_parse_extra_text_after_allele():
    parse("HLA-A*02:01 zipper")


def test_bad_input_parse_extra_text_after_allele_no_raise():
    result = parse("HLA-A*02:01 zipper", raise_on_error=False)
    assert result is None


def test_bad_input_parse_two_species():
    result = parse("HLA Calu", raise_on_error=False)
    assert result is None, result


def test_bad_input_only_MHC():
    result = parse("MHC", raise_on_error=False)
    assert result is None, result


def test_bad_input_only_MHC_in_three_words():
    result = parse("major histocompatibility complex", raise_on_error=False)
    assert result is None, result


def test_bad_input_only_numbers():
    result = parse("123", raise_on_error=False)
    assert result is None, result


# ---------------------------------------------------------------------------
# Null markers and punctuation-only strings
# https://github.com/pirl-unc/mhcgnomes/issues/102
#
# These are how a curator or an exported spreadsheet writes "missing". They
# used to fall through to the default species and come back as a confident
# result -- "-" as Homo sapiens, "n/a" as the rat haplotype RT1-n/A.
# ---------------------------------------------------------------------------

NULL_MARKERS = [
    "n/a",
    "N/A",
    "n/A",
    "N/a",
    "-",
    "--",
    "---",
    "\u2014",  # em dash
    "\u2013",  # en dash
    ".",
    "..",
    "...",
    "*",
    "**",
    ",",
    ";",
    ":",
    "|",
    "_",
    "/",
    "//",
    " - ",
]


@pytest.mark.parametrize("marker", NULL_MARKERS)
def test_null_marker_does_not_parse(marker):
    assert parse(marker, raise_on_error=False) is None


@pytest.mark.parametrize("marker", NULL_MARKERS)
def test_null_marker_raises_when_asked(marker):
    with pytest.raises(ParseError):
        parse(marker)


def test_punctuation_only_strings_do_not_parse():
    """
    An MHC name has to carry at least one letter or digit. 254 punctuation-only
    strings used to come back as Homo sapiens.
    """
    punctuation = "-_./\\|,;:*+#~ "
    for first in punctuation:
        for second in punctuation:
            token = first + second
            assert parse(token, raise_on_error=False) is None, token


# The null-marker guard must not swallow real haplotype shorthand, which is
# genuinely written with a slash and bare letters in the mouse literature.


def test_bare_haplotype_pair_still_parses():
    result = parse("b/d")
    eq_(type(result), Haplotype)
    eq_(result.name, "b/d")
    assert result.is_mouse


def test_bare_single_letter_haplotype_still_parses():
    eq_(parse("d").name, "d")
    eq_(parse("n").name, "n")


def test_bare_single_letter_gene_still_parses():
    eq_(parse("a").to_string(), "HLA-A")



import pytest

from mhcgnomes import parse

from .common import eq_


def test_parse_C0301_use_allele_aliases():
    allele = parse("C0301", use_allele_aliases=True)
    eq_(allele.to_string(), "HLA-C*03:04:01:01")


def test_parse_C0301_no_allele_aliases():
    allele = parse("C0301", use_allele_aliases=False)
    eq_(allele.to_string(), "HLA-C*03:01")


def test_exact_alias_can_rescue_an_otherwise_unparseable_legacy_name():
    allele = parse("Caja-PS*02:01", use_allele_aliases=True)

    assert allele.to_string() == "Caja-PS2*01"
    assert allele.raw_string == "Caja-PS*02:01"
    assert allele.is_pseudogene


def test_unparseable_legacy_alias_remains_opt_in():
    assert parse("Caja-PS*02:01", use_allele_aliases=False, raise_on_error=False) is None


@pytest.mark.parametrize(
    "near_miss",
    [
        "Caja-PS*02:02",
        "Caja-PS*03:01",
        "Caja-PS3*01",
        "MaLA-PS*02:01",
        "Capy-PS*02:01",
    ],
)
def test_caja_ps2_alias_does_not_generalize_to_near_misses_or_other_species(near_miss):
    assert parse(near_miss, use_allele_aliases=True, raise_on_error=False) is None



#!/usr/bin/env python3
"""
Extract MHC allele strings from a published paper's supplementary data.

Supports:
  - Excel files (.xlsx, .xls)
  - CSV/TSV files
  - Plain text (one allele per line or whitespace-separated)

The script scans all columns/cells for strings that look like MHC allele
names and outputs a validation TSV.

Usage:
    # Single file with known species
    python paper/scripts/scrape_paper.py \
        --input supp_table_S1.xlsx \
        --species "Parus major" \
        --source "PMID:12345678" \
        --output paper/validation/pmid_12345678.tsv

    # Auto-detect species from column headers
    python paper/scripts/scrape_paper.py \
        --input supp_table.csv \
        --source "doi:10.1234/example" \
        --output paper/validation/example.tsv

    # Batch: process all files in a directory
    python paper/scripts/scrape_paper.py \
        --input-dir paper/raw/pmid_12345678/ \
        --species "Gallus gallus" \
        --source "PMID:12345678" \
        --output paper/validation/pmid_12345678.tsv
"""

import argparse
import csv
import re
import sys
from pathlib import Path

# Patterns that look like MHC allele/gene names
MHC_PATTERNS = [
    # Standard allele: HLA-A*02:01, Mamu-B*017:01, BoLA-DRB3*01:01
    re.compile(
        r"\b[A-Z][A-Za-z0-9]{1,15}-[A-Za-z]{1,10}\*[\d:]+[A-Za-z]?\b"
    ),
    # Gene with prefix: HLA-A, Gaga-BF1, Dare-UBA, Modo-UA1
    re.compile(
        r"\b[A-Z][A-Za-z0-9]{1,15}-[A-Z][A-Za-z0-9]{0,8}\b"
    ),
    # Allele without prefix: A*02:01, DRB1*01:01, BF*21
    re.compile(r"\b[A-Z][A-Za-z0-9]{0,6}\*[\d:]+[A-Za-z]?\b"),
    # Mouse/rat style: H2-Kk, H2-Db, H-2Kb, RT1-Aa
    re.compile(r"\bH-?2-[A-Z][a-z0-9]*\b"),
    re.compile(r"\bRT1-[A-Za-z0-9.]+\b"),
    # Bird MHC with Mhc prefix: MhcPama-DAB1*01
    re.compile(r"\bMhc[A-Z][A-Za-z]+-[A-Za-z0-9*:]+\b"),
    # Non-mammalian: UAA*01, DAB*02:01, BLB1*04
    re.compile(r"\b(?:UA[A-Z]?|UB[A-Z]?|DA[AB]|DB[AB]|DC[AB]|DR[AB]|BLB|BF)\d?\*[\d:]+\b"),
    # Serotype-style: HLA-A2, HLA-B27, A2, B44
    re.compile(r"\bHLA-[A-Z]\d{1,3}\b"),
]

# Strings that match patterns but are NOT MHC names
FALSE_POSITIVE_PATTERNS = [
    re.compile(r"^\d+$"),  # bare numbers
    re.compile(r"^[A-Z]-\d+$"),  # plate coordinates like A-1
    re.compile(r"^rs\d+$"),  # SNP IDs
    re.compile(r"^chr\d", re.IGNORECASE),  # chromosome names
    re.compile(r"^ENSG\d"),  # Ensembl IDs
    re.compile(r"^LOC\d"),  # NCBI LOC IDs
    re.compile(r"^p\.\w+$"),  # protein changes
]


def looks_like_mhc(s):
    """Check if a string looks like an MHC allele/gene name."""
    s = s.strip()
    if len(s) < 3 or len(s) > 60:
        return False
    for fp in FALSE_POSITIVE_PATTERNS:
        if fp.match(s):
            return False
    for pat in MHC_PATTERNS:
        if pat.search(s):
            return True
    return False


def extract_from_text(text):
    """Extract MHC-like strings from free text."""
    candidates = set()
    for pat in MHC_PATTERNS:
        for match in pat.finditer(text):
            s = match.group()
            if not any(fp.match(s) for fp in FALSE_POSITIVE_PATTERNS):
                candidates.add(s)
    return candidates


def read_excel(path):
    """Read all sheets of an Excel file, return list of cell values."""
    try:
        import openpyxl
    except ImportError:
        print("Requires openpyxl: pip install openpyxl", file=sys.stderr)
        sys.exit(1)

    cells = []
    wb = openpyxl.load_workbook(path, read_only=True, data_only=True)
    for sheet in wb.sheetnames:
        ws = wb[sheet]
        for row in ws.iter_rows():
            for cell in row:
                if cell.value is not None:
                    cells.append(str(cell.value))
    wb.close()
    return cells


def read_csv_tsv(path):
    """Read a CSV or TSV file, return list of cell values."""
    cells = []
    with open(path, newline="", errors="replace") as f:
        # Sniff delimiter
        sample = f.read(4096)
        f.seek(0)
        if "\t" in sample:
            delimiter = "\t"
        elif "," in sample:
            delimiter = ","
        else:
            # Treat as one-per-line
            return [line.strip() for line in f if line.strip()]

        reader = csv.reader(f, delimiter=delimiter)
        for row in reader:
            cells.extend(row)
    return cells


def read_text(path):
    """Read a plain text file, return list of lines/words."""
    with open(path, errors="replace") as f:
        return [line.strip() for line in f if line.strip()]


def read_docx(path):
    """Read a .docx file and return paragraphs + table cells as text."""
    try:
        import zipfile
        import xml.etree.ElementTree as ET
    except ImportError:
        return []

    texts = []
    try:
        with zipfile.ZipFile(path) as z:
            # Read main document
            if "word/document.xml" in z.namelist():
                tree = ET.parse(z.open("word/document.xml"))
                ns = {"w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main"}
                for para in tree.iter("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}p"):
                    text = "".join(
                        node.text or ""
                        for node in para.iter("{http://schemas.openxmlformats.org/wordprocessingml/2006/main}t")
                    )
                    if text.strip():
                        texts.append(text.strip())
    except Exception:
        pass
    return texts


def process_file(path):
    """Read a file and extract MHC-like strings."""
    path = Path(path)
    suffix = path.suffix.lower()

    if suffix in (".xlsx", ".xls"):
        cells = read_excel(path)
    elif suffix in (".csv", ".tsv", ".txt"):
        cells = read_csv_tsv(path)
    elif suffix == ".docx":
        cells = read_docx(path)
    else:
        cells = read_text(path)

    # Strategy 1: check if whole cells are MHC names
    mhc_strings = set()
    for cell in cells:
        cell = cell.strip()
        if looks_like_mhc(cell):
            mhc_strings.add(cell)

    # Strategy 2: extract MHC patterns from longer text cells
    for cell in cells:
        if len(cell) > 20:  # likely a sentence or description
            mhc_strings.update(extract_from_text(cell))

    return sorted(mhc_strings)


def main():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("--input", help="Single input file")
    parser.add_argument("--input-dir", help="Directory of input files")
    parser.add_argument("--species", default="", help="Expected species (latin name)")
    parser.add_argument("--source", required=True, help="Source identifier (PMID or DOI)")
    parser.add_argument(
        "--output", default="-", help="Output TSV file (default: stdout)"
    )
    args = parser.parse_args()

    if not args.input and not args.input_dir:
        parser.error("Provide --input or --input-dir")

    # Collect input files
    input_files = []
    if args.input:
        input_files.append(Path(args.input))
    if args.input_dir:
        d = Path(args.input_dir)
        input_files.extend(
            sorted(
                p
                for p in d.iterdir()
                if p.suffix.lower() in (".xlsx", ".xls", ".csv", ".tsv", ".txt")
            )
        )

    if not input_files:
        print("No input files found.", file=sys.stderr)
        return 1

    # Extract from all files
    all_strings = set()
    for path in input_files:
        print(f"Processing {path}...", file=sys.stderr)
        strings = process_file(path)
        print(f"  Found {len(strings)} MHC-like strings", file=sys.stderr)
        all_strings.update(strings)

    print(f"\nTotal unique MHC strings: {len(all_strings)}", file=sys.stderr)

    # Write output
    out = sys.stdout if args.output == "-" else open(args.output, "w")
    writer = csv.DictWriter(
        out,
        fieldnames=["raw_string", "expected_species", "source"],
        delimiter="\t",
    )
    writer.writeheader()
    for s in sorted(all_strings):
        writer.writerow({
            "raw_string": s,
            "expected_species": args.species,
            "source": args.source,
        })

    if args.output != "-":
        out.close()
        print(f"Written to {args.output}", file=sys.stderr)

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Tests for 'alpha'/'beta' chain suffix resolution in gene names,
TL gene inheritance across Mus species, and literature-attested
Mus short prefixes.
"""

import pytest

from mhcgnomes import Gene, Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# alpha/beta suffix → single-letter chain indicator
# ---------------------------------------------------------------------------


def test_rt1_DOalpha_parses_as_gene():
    """'DOalpha' should resolve to gene DOa, not Allele(DOa, lpha)."""
    result = parse("RT1-DOalpha")
    assert type(result).__name__ == "Gene"
    eq_(result, Gene.get("RT1", "DOa"))


def test_rt1_DObeta_parses_as_gene():
    """'DObeta' should resolve to gene DOb."""
    result = parse("RT1-DObeta")
    assert type(result).__name__ == "Gene"
    eq_(result, Gene.get("RT1", "DOb"))


def test_mhc_rt1_dot_DOalpha():
    """Full benchmark string 'MHC RT1.DOalpha' should parse as Gene(DOa)."""
    result = parse("MHC RT1.DOalpha")
    assert type(result).__name__ == "Gene"
    eq_(result, Gene.get("RT1", "DOa"))


def test_h2_Ebeta_parses_as_gene():
    eq_(parse("H2-Ebeta"), Gene.get("H2", "EB"))


def test_h2_Aalpha_parses_as_gene():
    eq_(parse("H2-Aalpha"), Gene.get("H2", "AA"))


def test_exact_gene_match_not_broken_by_suffix_stripping():
    """Exact gene names ending in 'a' or 'b' should still match directly."""
    eq_(parse("RT1-DOa"), Gene.get("RT1", "DOa"))
    eq_(parse("HLA-DRA"), Gene.get("HLA", "DRA"))
    eq_(parse("HLA-DRB1"), Gene.get("HLA", "DRB1"))


def test_suffix_stripping_does_not_fire_when_candidate_not_a_gene():
    """When stripping 'alpha'/'beta' doesn't yield a known gene,
    the parser should fall through to other strategies, not crash."""
    # 'Ka' is not a Mus gene, so H2-Kalpha should not parse as Gene(Ka)
    result = parse("H2-Kalpha", raise_on_error=False)
    if result is not None:
        assert result.name != "Ka"


# ---------------------------------------------------------------------------
# TL gene on Mus sp.
# ---------------------------------------------------------------------------


def test_tl_gene_defined_on_mus_sp():
    gene = Gene.get("MusSp", "TL")
    assert gene is not None


def test_muco_tl_parses():
    """MucoTL = Mus cookii prefix + TL gene."""
    eq_(parse("MucoTL"), Gene.get("Muco", "TL"))


def test_mupl_tl_parses():
    """MuplTL = Mus platythrix prefix + TL gene."""
    eq_(parse("MuplTL"), Gene.get("Mupl", "TL"))


def test_muca_tl_parses():
    eq_(parse("MucaTL"), Gene.get("Muca", "TL"))


def test_muab_tl_parses():
    eq_(parse("MuabTL"), Gene.get("Muab", "TL"))


def test_mumi_tl_parses():
    eq_(parse("MumiTL"), Gene.get("Mumi", "TL"))


def test_mupa_tl_parses():
    eq_(parse("MupaTL"), Gene.get("Mupa", "TL"))


def test_tl_inherited_by_mus_musculus():
    """H2 (Mus musculus) should also inherit TL from Mus sp."""
    gene = Gene.get("H2", "TL")
    assert gene is not None


# ---------------------------------------------------------------------------
# Literature-attested Mus short prefixes
# ---------------------------------------------------------------------------

ATTESTED_MUS_PREFIXES = [
    ("Muco", "Mus cookii"),
    ("Mupl", "Mus platythrix"),
    ("Muce", "Mus cervicolor"),
    ("Muca", "Mus caroli"),
    ("Mupa", "Mus pahari"),
    ("Muab", "Mus abbotti"),
    ("Mumi", "Mus minutoides"),
]


@pytest.mark.parametrize("prefix, latin", ATTESTED_MUS_PREFIXES)
def test_mus_short_prefix_resolves(prefix, latin):
    species = Species.get(prefix)
    assert species is not None, f"Prefix {prefix} not found"
    assert species.latin_name == latin


def test_musa_not_registered():
    """Musa (M. saxicola) is NOT attested in the literature and should
    not resolve — it collides with the banana genus Musa."""
    species = Species.get("Musa")
    if species is not None:
        assert species.latin_name != "Mus saxicola"
